from django import forms
from django.forms import inlineformset_factory, widgets
from django.contrib.auth.models import User, Group
from django.core.exceptions import ValidationError
from django.utils.translation import ugettext_lazy as _
from django.db.models import Q

from datetime import datetime, timedelta
from dateutil.relativedelta import relativedelta
from crispy_forms.helper import FormHelper
from crispy_forms.layout import Layout, Field, Div, Submit
from crispy_forms.bootstrap import Accordion, AccordionGroup

from danceschool.core.models import EventOccurrence, Event
from danceschool.core.utils.timezone import ensure_timezone
from danceschool.core.forms import LocationWithDataWidget

from .models import PrivateEvent, EventReminder

EVENT_REPEAT_CHOICES = [
    (str(timedelta(hours=1).total_seconds()),_('Hour')),
    (str(timedelta(days=1).total_seconds()),_('Day'),),
    (str(timedelta(days=7).total_seconds()),_('Week')),
    ('months=+1',_('Month')),
    ('years=+1',_('Year')),
]

VISIBILITY_CHOICES = [
    ('all',_('All Staff')),
    ('me',_('Only To Me')),
    ('group',_('Select User Group')),
    ('users',_('Select Users')),
]

REMINDER_SET_CHOICES = [
    ('none',_('No One')),
    ('all',_('All Staff')),
    ('me',_('Only To Me')),
    ('group',_('Select User Group')),
    ('users',_('Select Users')),
]

REMINDER_TIME_CHOICES = [
    (0,_('0 minutes')),
    (30,_('30 minutes')),
    (60, _('1 hour')),
    (720,_('12 hours')),
    (4320, _('3 days')),
]


class AddPrivateEventForm(forms.ModelForm):

    visibleTo = forms.ChoiceField(label=_('Make this event visible to:'),choices=VISIBILITY_CHOICES,initial='all')

    def __init__(self, *args, **kwargs):
        user = kwargs.pop('user', None)

        if user:
            kwargs.update(initial={
                'displayToUsers': [user.id,],
                'submissionUser': user.id,
            })

        super(AddPrivateEventForm,self).__init__(*args,**kwargs)
        self.fields['submissionUser'].widget = forms.HiddenInput()
        self.fields['status'].widget = forms.HiddenInput()
        self.fields['status'].initial = Event.RegStatus.hidden
        self.helper = FormHelper()
        self.helper.form_method = 'post'
        self.helper.form_tag = False  # Our template must explicitly include the <form tag>
        self.helper.layout = Layout(
            'status',
            'submissionUser',
            'title',
            Div(
                Field('category', wrapper_class='col'),
                Field('visibleTo', wrapper_class='col'),
                css_class='form-row'),
            Div('displayToUsers'),
            Div('displayToGroup'),
            Accordion(
                AccordionGroup(_('Add A Description'),'descriptionField',active=False),
                AccordionGroup(_('Add A Location'),Div('location','room','locationString')),
                AccordionGroup(_('Add a Link'),'link'),
            ),
        )

    def clean(self):
        super(AddPrivateEventForm,self).clean()
        visibleTo = self.cleaned_data.get('visibleTo')
        submissionUser = self.cleaned_data.get('submissionUser')

        # Set visibility based on the choice the user made
        if visibleTo == 'all':
            self.cleaned_data.pop('displayToUsers',None)
            self.cleaned_data.pop('displayToGroup',None)
        elif visibleTo == 'me':
            self.cleaned_data['displayToUsers'] = [submissionUser,]
            self.cleaned_data.pop('displayToGroup',None)
        elif visibleTo == 'users':
            self.cleaned_data.pop('displayToGroup',None)
        elif visibleTo == 'group':
            self.cleaned_data.pop('displayToUsers',None)

    class Meta:
        model = PrivateEvent
        exclude = []
        widgets = {
            'location': LocationWithDataWidget,
        }

    class Media:
        js = ('timepicker/jquery.timepicker.min.js','jquery-ui/jquery-ui.min.js','moment/moment.min.js','js/addEvent_rollups.js','js/serieslocation_capacity_change.js',)
        css = {'all':('timepicker/jquery.timepicker.css','jquery-ui/jquery-ui.min.css',)}


class AddEventOccurrenceForm(forms.ModelForm):
    allDay = forms.BooleanField(label=_('All Day'),required=False)

    def clean(self):
        startTime = self.cleaned_data['startTime']
        endTime = self.cleaned_data['endTime']
        allDay = self.cleaned_data.pop('allDay', None) or False
        if allDay:
            self.cleaned_data['startTime'] = ensure_timezone(datetime(startTime.year,startTime.month,startTime.day))
            self.cleaned_data['endTime'] = ensure_timezone(datetime(endTime.year,endTime.month,endTime.day)) + timedelta(days=1)
        if self.cleaned_data['endTime'] < self.cleaned_data['startTime']:
            raise ValidationError(_('End time cannot be before start time.'),code='invalid')

    class Meta:
        model = EventOccurrence
        exclude = []


class OccurrenceFormSetHelper(FormHelper):

    def __init__(self, *args, **kwargs):
        super(OccurrenceFormSetHelper, self).__init__(*args, **kwargs)

        self.form_tag = False  # Our template must explicitly include the <form tag>
        self.layout = Layout(
            Div(
                Field('startTime', wrapper_class='col',),
                Field('endTime', wrapper_class='col',),
                Field('allDay', wrapper_class='col',),
                css_class='form-row'
            ),
            Div(
                Field('extraOccurrencesToAdd', wrapper_class='col',),
                Field('extraOccurrenceRule', wrapper_class='col',),
                css_class="form-row"
            ),
            Div(
                Field('sendReminderTo', wrapper_class='col',),
                Field('sendReminderWhen', wrapper_class='col',),
                Field('sendReminderWhich', wrapper_class='col',),
                css_class="form-row"
            ),
            Div('sendReminderGroup'),
            Div('sendReminderUsers'),
            Submit('submit', u'Submit', css_class='btn btn-primary'),
        )


class EventOccurrenceCustomFormSet(forms.BaseInlineFormSet):
    ''' Formset for occurrences added via the Private Events form '''

    def add_fields(self, form, index):
        super(EventOccurrenceCustomFormSet,self).add_fields(form,index)
        form.fields['startTime'] = forms.SplitDateTimeField(label=_("Start Time"),input_time_formats=['%I:%M%p','%-I:%M%p'], widget=widgets.SplitDateTimeWidget)
        form.fields['endTime'] = forms.SplitDateTimeField(label=_("End Time"),input_time_formats=['%I:%M%p','%-I:%M%p'], widget=widgets.SplitDateTimeWidget)
        form.fields['extraOccurrencesToAdd'] = forms.IntegerField(label=_("Repeat __ times:"),initial=0,min_value=0,max_value=100,required=False)
        form.fields['extraOccurrenceRule'] = forms.ChoiceField(label=_("Every:"),choices=EVENT_REPEAT_CHOICES,required=False)
        form.fields['sendReminderTo'] = forms.ChoiceField(label=_("Send A Reminder To:"),choices=REMINDER_SET_CHOICES,initial='none',required=False)
        form.fields['sendReminderWhen'] = forms.ChoiceField(label=_("Before:"),choices=REMINDER_TIME_CHOICES,initial=0,required=False)
        form.fields['sendReminderWhich'] = forms.ChoiceField(label=_("Ahead Of:"),choices=(('all',_('All Occurrences')),('first',_('First Occurrence Only'))),initial='all',required=False)
        form.fields['sendReminderGroup'] = forms.ModelChoiceField(label=_("Remind Group:"),queryset=Group.objects.all(),required=False)
        form.fields['sendReminderUsers'] = forms.ModelMultipleChoiceField(label=_("Remind Users:"),queryset=User.objects.filter(staffmember__isnull=False),required=False)

    def setReminder(self,occurrence):
        '''
        This function is called to create the actual reminders for each occurrence that is created.
        '''
        sendReminderTo = self[0].cleaned_data['sendReminderTo']
        sendReminderWhen = self[0].cleaned_data['sendReminderWhen']
        sendReminderGroup = self[0].cleaned_data['sendReminderGroup']
        sendReminderUsers = self[0].cleaned_data['sendReminderUsers']

        # Set the new reminder's time
        new_reminder_time = occurrence.startTime - timedelta(minutes=int(float(sendReminderWhen)))
        new_reminder = EventReminder(eventOccurrence=occurrence,time=new_reminder_time)
        new_reminder.save()

        # Set reminders based on the choice the user made
        if sendReminderTo == 'all':
            user_set = User.objects.filter(Q(staffmember__isnull=False) | Q(is_staff=True))
        elif sendReminderTo == 'me':
            user_set = User.objects.filter(id=occurrence.event.submissionUser.id)
        elif sendReminderTo == 'users':
            user_set = User.objects.filter(**{'id__in': sendReminderUsers})
        elif sendReminderTo == 'group':
            user_set = User.objects.filter(**{'groups': sendReminderGroup})
        else:
            user_set = []

        for user in user_set:
            new_reminder.notifyList.add(user)

    def save(self):
        extraOccurrencesToAdd = self[0].cleaned_data['extraOccurrencesToAdd']
        extraOccurrenceRule = self[0].cleaned_data['extraOccurrenceRule']
        sendReminderTo = self[0].cleaned_data['sendReminderTo']
        sendReminderWhich = self[0].cleaned_data['sendReminderWhich']

        # Save this actual occurrence first
        this_occurrence = super(EventOccurrenceCustomFormSet, self).save()

        # Create a reminder for this occurrence if applicable
        if sendReminderTo in ['all','me','users','group']:
            self.setReminder(this_occurrence[0])

        # Create new occurrences based on the number of occurrences to add and the occurrence rule given
        remainingOccurrences = extraOccurrencesToAdd

        # These will be repopulated with each loop and used to calculate the next event times
        old_startTime = this_occurrence[0].startTime
        old_endTime = this_occurrence[0].endTime

        # Make a new occurrence for each requested new occurrence.  Also, make a reminder if requested
        while remainingOccurrences > 0:
            if 'months' not in extraOccurrenceRule and 'years' not in extraOccurrenceRule:
                time_diff = timedelta(seconds=float(extraOccurrenceRule))
                new_startTime = old_startTime + time_diff
                new_endTime = old_endTime + time_diff
            elif 'months' in extraOccurrenceRule:
                num_months = int(float(extraOccurrenceRule.replace('months=','')))
                new_startTime = old_startTime + relativedelta(months=num_months)
                new_endTime = old_endTime + relativedelta(months=num_months)
            elif 'years' in extraOccurrenceRule:
                num_years = int(float(extraOccurrenceRule.replace('years=','')))
                new_startTime = old_startTime + relativedelta(years=num_years)
                new_endTime = old_endTime + relativedelta(years=num_years)

            # Create the occurrence
            new_occ = EventOccurrence(event=this_occurrence[0].event, startTime=new_startTime,endTime=new_endTime)
            new_occ.save()

            # Create the reminders
            if sendReminderWhich == 'all' and sendReminderTo in ['all','me','users','group']:
                self.setReminder(new_occ)

            # Finally, prepare for the next iteration of the loop
            old_startTime = new_startTime
            old_endTime = new_endTime
            remainingOccurrences -= 1


EventOccurrenceFormSet = inlineformset_factory(
    PrivateEvent, EventOccurrence,
    formset=EventOccurrenceCustomFormSet,
    form=AddEventOccurrenceForm,
    exclude=[],
    extra=1, can_delete=False
)



'''
This file defines a variety of preferences that must be set in the DB,
but can be changed dynamically.
'''

from django.utils.translation import ugettext_lazy as _
from django.template.loader import get_template

from dynamic_preferences.types import BooleanPreference, StringPreference, IntegerPreference, FloatPreference, ChoicePreference, ModelChoicePreference, Section
from dynamic_preferences.registries import global_preferences_registry
from filer.models import Folder
from cms.models import Page
from cms.forms.fields import PageSelectFormField

from .utils.serializers import PageModelSerializer
from .models import EventStaffCategory, EmailTemplate, get_defaultEmailName, get_defaultEmailFrom

# we create some section objects to link related preferences together

general = Section('general', _('General Settings'))
contact = Section('contact', _('Contact Info'))
registration = Section('registration', _('Registration'))
email = Section('email', _('Email'))
calendar = Section('calendar', _('Calendar'))


############################
# General Preferences
#
@global_preferences_registry.register
class CurrencyCode(StringPreference):
    section = general
    name = 'currencyCode'
    verbose_name = _('Currency Code')
    help_text = _('Use a three-letter currency code for compatibility with payment systems')
    default = 'USD'


@global_preferences_registry.register
class CurrencySymbol(StringPreference):
    section = general
    name = 'currencySymbol'
    verbose_name = _('Currency Symbol')
    help_text = _('Enter a symbol to use on the website representing the currency used (e.g. $)')
    default = '$'


@global_preferences_registry.register
class StaffPhotosFolder(ModelChoicePreference):
    section = general
    model = Folder
    queryset = Folder.objects.all()
    name = 'staffPhotosFolder'
    verbose_name = _('Staff Photos Folder')
    help_text = _(
        'The folder used to store profile photos of staff members. ' +
        'Note that if this is changed, existing profile photos linked to staff members will not be moved.')
    default = Folder.objects.none()


@global_preferences_registry.register
class DefaultAdminSuccessPage(IntegerPreference):
    section = general
    model = Page
    name = 'defaultAdminSuccessPage'
    verbose_name = _('Default Admin Form Success Page')
    help_text = _('The page to which a staff user is redirected after successfully submitting an admin form.')
    default = Page.objects.none()
    field_class = PageSelectFormField

    def __init__(self, *args, **kwargs):
        ''' Changes the default serializer '''
        super(self.__class__, self).__init__(*args, **kwargs)
        self.serializer = PageModelSerializer


@global_preferences_registry.register
class EnableCronTasks(BooleanPreference):
    section = general
    name = 'enableCronTasks'
    verbose_name = _('Enable Periodic (Cron) Tasks')
    help_text = _(
        'Uncheck this if you will be using crontab, a scheduling service such as Heroku, ' +
        'or if you otherwise do not wish to permit automatic generation of expenses, closing ' +
        'of class registration, etc.'
    )
    default = True


@global_preferences_registry.register
class StaffCategoryInstructor(ModelChoicePreference):
    section = general
    name = 'eventStaffCategoryInstructor'
    verbose_name = _('Instructor Event Staff Category')
    model = EventStaffCategory
    queryset = EventStaffCategory.objects.all()

    def get_default(self):
        # if self.model and self.model._meta.db_table in connection.introspection.table_names():
        return EventStaffCategory.objects.get_or_create(name=_('Class Instruction'))[0]


@global_preferences_registry.register
class StaffCategoryAssistant(ModelChoicePreference):
    section = general
    name = 'eventStaffCategoryAssistant'
    verbose_name = _('Assistant Instructor Event Staff Category')
    model = EventStaffCategory
    queryset = EventStaffCategory.objects.all()

    def get_default(self):
        # if self.model and self.model._meta.db_table in connection.introspection.table_names():
        return EventStaffCategory.objects.get_or_create(name=_('Assistant Class Instruction'))[0]


@global_preferences_registry.register
class StaffCategorySubstitute(ModelChoicePreference):
    section = general
    name = 'eventStaffCategorySubstitute'
    verbose_name = _('Substitute Teacher Event Staff Category')
    model = EventStaffCategory
    queryset = EventStaffCategory.objects.all()

    def get_default(self):
        # if self.model and self.model._meta.db_table in connection.introspection.table_names():
        return EventStaffCategory.objects.get_or_create(name=_('Substitute Teaching'))[0]


@global_preferences_registry.register
class StaffCategoryOther(ModelChoicePreference):
    section = general
    name = 'eventStaffCategoryOther'
    verbose_name = _('Other Staff Event Staff Category')
    model = EventStaffCategory
    queryset = EventStaffCategory.objects.all()

    def get_default(self):
        # if self.model and self.model._meta.db_table in connection.introspection.table_names():
        return EventStaffCategory.objects.get_or_create(name=_('Other Staff'))[0]


#################################
# Contact Information Preferences
#
@global_preferences_registry.register
class BusinessName(StringPreference):
    section = contact
    name = 'businessName'
    verbose_name = _('Business Name')
    help_text = _('For invoices and template use')
    default = ''


@global_preferences_registry.register
class BusinessPhone(StringPreference):
    section = contact
    name = 'businessPhone'
    verbose_name = _('Business phone number')
    default = ''

    def get_field_kwargs(self):
        field_kwargs = super(self.__class__,self).get_field_kwargs()
        field_kwargs['required'] = False
        return field_kwargs


@global_preferences_registry.register
class BusinessEmail(StringPreference):
    section = contact
    name = 'businessEmail'
    verbose_name = _('Business email address')
    help_text = 'For invoices and template use'
    default = ''


@global_preferences_registry.register
class BusinessAddress(StringPreference):
    section = contact
    name = 'businessAddress'
    verbose_name = _('Business street/mailing address (Line 1)')
    default = ''

    def get_field_kwargs(self):
        field_kwargs = super(self.__class__,self).get_field_kwargs()
        field_kwargs['required'] = False
        return field_kwargs


@global_preferences_registry.register
class BusinessAddressLineTwo(StringPreference):
    section = contact
    name = 'businessAddressLineTwo'
    verbose_name = _('Business Address Line 2')
    help_text = 'Optional'
    default = ''

    def get_field_kwargs(self):
        field_kwargs = super(self.__class__,self).get_field_kwargs()
        field_kwargs['required'] = False
        return field_kwargs


@global_preferences_registry.register
class BusinessCity(StringPreference):
    section = contact
    name = 'businessCity'
    verbose_name = _('City')
    default = ''

    def get_field_kwargs(self):
        field_kwargs = super(self.__class__,self).get_field_kwargs()
        field_kwargs['required'] = False
        return field_kwargs


@global_preferences_registry.register
class BusinessState(StringPreference):
    section = contact
    name = 'businessState'
    verbose_name = _('State')
    default = ''

    def get_field_kwargs(self):
        field_kwargs = super(self.__class__,self).get_field_kwargs()
        field_kwargs['required'] = False
        return field_kwargs


@global_preferences_registry.register
class BusinessZip(StringPreference):
    section = contact
    name = 'businessZip'
    verbose_name = _('ZIP/Postal Code')
    default = ''

    def get_field_kwargs(self):
        field_kwargs = super(self.__class__,self).get_field_kwargs()
        field_kwargs['required'] = False
        return field_kwargs


@global_preferences_registry.register
class BusinessCountryCode(StringPreference):
    section = contact
    name = 'businessCountryCode'
    verbose_name = _('Country code')
    help_text = _('Use a two-letter country code for compatibility with payment systems')
    default = 'US'


#################################
# Registration-Related Preferences
#
@global_preferences_registry.register
class EnableRegistration(BooleanPreference):
    section = registration
    name = 'registrationEnabled'
    verbose_name = _('Registration Enabled')
    help_text = _('Uncheck this to disable registration entirely.  Useful when performing upgrades or testing.')
    default = True


@global_preferences_registry.register
class DefaultClassCapacity(IntegerPreference):
    section = registration
    name = 'defaultEventCapacity'
    verbose_name = _('Default Registration Capacity')
    help_text = _('This is usually a location-specific setting or an event-specific setting.  However, if a limit is not otherwise provided, then this will limit the number of registrations allowed.')
    default = 50


@global_preferences_registry.register
class RegistrationClosesAfterDays(IntegerPreference):
    section = registration
    name = 'closeAfterDays'
    verbose_name = _('Close Registration Days After Beginning')
    help_text = _('By default, online registration closes automatically this many days from the beginning of an Event or Series (e.g. Enter 2 to close two days after beginning, or enter -3 to close three days before beginning).  This value can be overridden for individual Events.')
    default = 7


@global_preferences_registry.register
class RegistrationDisplayLimitDays(IntegerPreference):
    section = registration
    name = 'displayLimitDays'
    verbose_name = _('Do Not Show Registration for Events that Begin More Than __ Days in the Future')
    help_text = _('By default, events are shown on the registration page as soon as they are created.  If you list events far into the future, set this value, and events that begin more than this many days in the future will not be shown in the registration page.  Enter 0 for no restriction.')
    default = 0


@global_preferences_registry.register
class EventMonthRule(ChoicePreference):
    section = registration
    name = 'eventMonthRule'
    choices = [
        ('1', 'Month of first occurrence'),
        ('2', 'Month of second occurrence'),
        ('Last', 'Month of last occurrence'),
        ('FirstMulti', 'First month with more than one occurrence (or first)'),
        ('Most', 'Month with the most occurrences'),
    ]
    verbose_name = _('Rule for Assigning Events to Months')
    default = 'FirstMulti'
    help_text = _(
        'When events are listed by month in various views, this rule determines '
        'the month to which each event is assigned.'
    )


@global_preferences_registry.register
class SalesTaxRate(FloatPreference):
    section = registration
    name = 'salesTaxRate'
    verbose_name = _('Sales tax percentage rate to be applied registrations')
    help_text = _('Enter, e.g. \'10\' for a 10% tax rate to be applied to all class and event registrations.')
    default = 0.0


@global_preferences_registry.register
class BuyerPaysSalesTax(BooleanPreference):
    section = registration
    name = 'buyerPaysSalesTax'
    verbose_name = _('Buyer pays sales tax (added to total price)')
    help_text = _('If unchecked, then the buyer will not be charged sales tax directly, but the amount of tax collected by the business will be reported.')
    default = True


@global_preferences_registry.register
class AllowAjaxSignin(BooleanPreference):
    section = registration
    name = 'allowAjaxSignin'
    verbose_name = _('Allow users to login/signup during registration')
    help_text = _('If you don\'t allow customers to see their history or do automatic checks for prerequisites, then you may not need customers to create user accounts or authenticate themselves. If this box is unchecked, then users will not be able to log in or create a new account during the registration process.')
    default = True


@global_preferences_registry.register
class RegistrationSessionExpiryMinutes(IntegerPreference):
    section = registration
    name = 'sessionExpiryMinutes'
    verbose_name = _('Temporary registration data expires after __ minutes')
    help_text = _('In each step of the registration process, customers have this many minutes to proceed before they are no longer guaranteed a spot and are required to begin again.')
    default = 15


@global_preferences_registry.register
class DeleteExpiredTemporaryRegistrations(BooleanPreference):
    section = registration
    name = 'deleteExpiredTemporaryRegistrations'
    verbose_name = _('Automatically delete expired temporary registration and session data')
    help_text = _('If this box is checked, then an hourly script will automatically expired temporary registration and session data.  Disabling this feature is only recommended for testing.')
    default = True


############################
# Email Preferences
#
@global_preferences_registry.register
class SiteEmailsDisabled(BooleanPreference):
    section = email
    name = 'disableSiteEmails'
    verbose_name = _('Disable Sending of Emails')
    help_text = _('Check to disable all standard emails from being sent.')
    default = False


@global_preferences_registry.register
class DefaultEmailsFrom(StringPreference):
    section = email
    name = 'defaultEmailFrom'
    verbose_name = _('Email address that general site emails are sent from')
    help_text = _('This email address may be overridden by specific email templates')
    default = ''

    def get_field_kwargs(self):
        field_kwargs = super(self.__class__,self).get_field_kwargs()
        field_kwargs['required'] = False
        return field_kwargs


@global_preferences_registry.register
class DefaultEmailsName(StringPreference):
    section = email
    name = 'defaultEmailName'
    verbose_name = _('Name that is used for general site emails')
    help_text = _('This name may be overridden by specific email templates')
    default = ''

    def get_field_kwargs(self):
        field_kwargs = super(self.__class__,self).get_field_kwargs()
        field_kwargs['required'] = False
        return field_kwargs


@global_preferences_registry.register
class ErrorEmailsEnabled(BooleanPreference):
    section = email
    name = 'enableErrorEmails'
    verbose_name = _('Send Error Emails')
    default = True


@global_preferences_registry.register
class ErrorEmailsFrom(StringPreference):
    section = email
    name = 'errorEmailFrom'
    verbose_name = _('Email address that error emails are sent from')
    default = ''

    def get_field_kwargs(self):
        field_kwargs = super(self.__class__,self).get_field_kwargs()
        field_kwargs['required'] = False
        return field_kwargs


@global_preferences_registry.register
class ErrorEmailsTo(StringPreference):
    section = email
    name = 'errorEmailTo'
    verbose_name = _('Email address that error emails are sent to')
    default = ''

    def get_field_kwargs(self):
        field_kwargs = super(self.__class__,self).get_field_kwargs()
        field_kwargs['required'] = False
        return field_kwargs


@global_preferences_registry.register
class EmailLinkProtocol(ChoicePreference):
    section = email
    name = 'linkProtocol'
    choices = [
        ('http','HTTP'),
        ('https','HTTPS'),
    ]
    verbose_name = _('Protocol for URLs in email templates')
    default = 'http'
    help_text = _(
        'Invoice emails and other emails sometimes include links that require ' +
        'a protocol to be specified via the "protocol" context variable.  If ' +
        'your site uses solely HTTPS, you may need to change this value to avoid ' +
        'broken links.'
    )


@global_preferences_registry.register
class RegSuccessEmailTemplate(ModelChoicePreference):
    section = email
    name = 'registrationSuccessTemplate'
    verbose_name = _('Email template used for successful email registrations')
    model = EmailTemplate
    queryset = EmailTemplate.objects.all()

    def get_default(self):
        # if self.model and self.model._meta.db_table in connection.introspection.table_names():

        initial_template = get_template('email/registration_success.html')
        with open(initial_template.origin.name,'r') as infile:
            content = infile.read()
            infile.close()

        return EmailTemplate.objects.get_or_create(
            name=_('Registration Confirmation Email'),
            defaults={
                'subject': _('Registration Confirmation'),
                'content': content or '',
                'defaultFromAddress': get_defaultEmailFrom(),
                'defaultFromName': get_defaultEmailName(),
                'defaultCC': '',
                'hideFromForm': True,
            }
        )[0]


@global_preferences_registry.register
class InvoiceEmailTemplate(ModelChoicePreference):
    section = email
    name = 'invoiceTemplate'
    verbose_name = _('Email template used for invoice email creation')
    model = EmailTemplate
    queryset = EmailTemplate.objects.all()

    def get_default(self):
        # if self.model and self.model._meta.db_table in connection.introspection.table_names():

        initial_template = get_template('email/invoice_initial.html')
        with open(initial_template.origin.name,'r') as infile:
            content = infile.read()
            infile.close()

        return EmailTemplate.objects.get_or_create(
            name=_('Registration Invoice Email'),
            defaults={
                'subject': _('Registration Invoice'),
                'content': content or '',
                'defaultFromAddress': get_defaultEmailFrom(),
                'defaultFromName': get_defaultEmailName(),
                'defaultCC': '',
                'hideFromForm': True,
            }
        )[0]


############################
# Calendar Preferences
#
@global_preferences_registry.register
class EnableCalendarFeed(BooleanPreference):
    section = calendar
    name = 'calendarFeedEnabled'
    verbose_name = _('Enable public calendar feed')
    help_text = _('Uncheck to disable the automatic creation of a public calendar feed.')
    default = True


@global_preferences_registry.register
class EnablePrivateCalendarFeed(BooleanPreference):
    section = calendar
    name = 'privateCalendarFeedEnabled'
    verbose_name = _('Enable private calendar feed')
    help_text = _('Uncheck to disable the automatic creation of private (per-instructor) calendar feeds.')
    default = True


@global_preferences_registry.register
class CalendarClassColor(StringPreference):
    section = calendar
    name = 'defaultClassColor'
    verbose_name = _('Default color used for classes on the calendar')
    help_text = _('Enter a CSS compatible color')
    default = '#477087'


@global_preferences_registry.register
class CalendarEventColor(StringPreference):
    section = calendar
    name = 'defaultEventColor'
    verbose_name = _('Default color used for events on the calendar')
    help_text = _('Enter a CSS compatible color')
    default = '#477087'



from django import forms
from django.conf import settings
from django.contrib.auth.models import User
from django.core.validators import ValidationError
from django.core.urlresolvers import reverse
from django.db.models import Q
from django.forms.widgets import Select
from django.utils.encoding import force_text
from django.utils.safestring import mark_safe
from django.utils.html import format_html
from django.utils.translation import ugettext_lazy as _
from django.utils import timezone

from crispy_forms.helper import FormHelper
from crispy_forms.layout import Layout, Field, Div, Submit, HTML
from collections import OrderedDict
from dal import autocomplete
import logging

from danceschool.core.models import InvoiceItem

from .models import ExpenseItem, ExpenseCategory, RevenueItem, StaffMemberWageInfo
from .autocomplete_light_registry import get_method_list


# Define logger for this file
logger = logging.getLogger(__name__)


RECIPIENT_CHOICES = (
                    (1,_('Me')),
                    (2,_('A Location (e.g. a rental payment)')),
                    (3,_('Someone else')),)

PAYBY_CHOICES = (
                (1,_('Hours of Work/Rental (paid at default rate)')),
                (2,_('Flat Payment')),)

REVENUE_ASSOCIATION_CHOICES = (
                              (2, _('A Class Series or Event')),
                              (3, _('Neither')),)


class ExpenseCategoryWidget(Select):
    '''
    Override render_option to permit extra data of default wage to be used by JQuery
    This could be optimized to reduce database calls by overriding the render function.
    '''

    def render_option(self, selected_choices, option_value, option_label):
        if option_value is None:
            option_value = ''
        option_value = force_text(option_value)
        if option_value in selected_choices:
            selected_html = mark_safe(' selected="selected"')
            if not self.allow_multiple_selected:
                # Only allow for a single selection.
                selected_choices.remove(option_value)
        else:
            selected_html = ''

        # Pass the default wage rate as an option
        if option_value:
            defaultRate = ExpenseCategory.objects.filter(id=int(option_value)).first().defaultRate
            extra_value_data = ' data-defaultRate=' + str(defaultRate)
        else:
            extra_value_data = ''

        return format_html('<option value="{}"{}{}>{}</option>',
                           option_value,
                           selected_html,
                           extra_value_data,
                           force_text(option_label))


class ExpenseReportingForm(forms.ModelForm):
    payTo = forms.ChoiceField(widget=forms.RadioSelect, choices=RECIPIENT_CHOICES,label=_('This expense to be paid to:'),initial=1)
    payBy = forms.ChoiceField(widget=forms.RadioSelect, choices=PAYBY_CHOICES, label=_('Report this expense as:'), initial=1)
    paymentMethod = autocomplete.Select2ListCreateChoiceField(
        choice_list=get_method_list,
        required=False,
        widget=autocomplete.ListSelect2(url='paymentMethod-list-autocomplete'),
        label=_('Payment method'),
    )

    def __init__(self, *args, **kwargs):
        user = kwargs.pop('user', None)
        user_id = getattr(user,'id')

        super(ExpenseReportingForm,self).__init__(*args,**kwargs)

        self.helper = FormHelper()

        if user.has_perm('financial.mark_expenses_paid'):
            payment_section = Div(
                Div(
                    HTML('<a data-toggle="collapse" href="#collapsepayment">%s</a> (%s)' % (_('Mark as Approved/Paid'),_('click to expand'))),
                    css_class='card-header'
                ),
                Div(
                    'approved',
                    'paid',
                    Div(
                        Field('paymentDate', css_class='datepicker', wrapper_class='col-md-3'),
                        Field('paymentMethod', wrapper_class='col-md-6'),
                        css_class='form-row',
                    ),
                    # The hidden input of accrual date must be passed as a naive datetime.
                    # Django will take care of converting it to local time
                    Field('accrualDate', type="hidden",value=timezone.make_naive(timezone.now()) if timezone.is_aware(timezone.now()) else timezone.now()),
                    HTML('<p style="margin-top: 30px;"><strong>%s</strong> %s</p>' % (_('Note:'),_('For accounting purposes, please do not mark expenses as paid unless they have already been paid to the recipient.'))),
                    css_class='card-body collapse',
                    id='collapsepayment',
                ),
                css_class='card my-4')
        else:
            payment_section = None

        # Add category button should only appear for users who are allowed to add categories
        if user.has_perm('financial.add_expensecategory'):
            related_url = reverse('admin:financial_expensecategory_add') + '?_to_field=id&_popup=1'
            added_html = [
                '<a href="%s" class="btn btn-outline-secondary related-widget-wrapper-link add-related" id="add_id_category"> ' % related_url,
                '<img src="%sadmin/img/icon-addlink.svg" width="10" height="10" alt="%s"/></a>' % (getattr(settings,'STATIC_URL','/static/'), _('Add Another'))
            ]
            category_field = Div(
                Div('category',css_class='col-sm-11'),
                Div(HTML('\n'.join(added_html)),css_class='col-sm-1',style='margin-top: 25px;'),
                css_class='related-widget-wrapper row'
            )
        else:
            category_field = Div('category')

        self.helper.layout = Layout(
            Field('submissionUser', type="hidden", value=user_id),
            Field('payToUser', type="hidden", value=user_id),
            'payTo',
            'payBy',
            'payToLocation',
            'payToName',
            category_field,
            'description',
            'hours',
            'total',
            'reimbursement',
            payment_section,
            'attachment',
            Submit('submit',_('Submit')),
        )

    def clean(self):
        # Custom cleaning ensures that user, hours, and total
        # are not reported where not necessary.
        super(ExpenseReportingForm, self).clean()

        payTo = self.cleaned_data.get('payTo')
        payToUser = self.cleaned_data.get('payToUser')
        payToLocation = self.cleaned_data.get('payToLocation')
        payToName = self.cleaned_data.get('payToName')
        payBy = self.cleaned_data.get('payBy')
        hours = self.cleaned_data.get('hours')
        total = self.cleaned_data.get('total')

        paid = self.cleaned_data.get('paid')
        paymentDate = self.cleaned_data.get('paymentDate')

        # Automatically marks expenses that are paid
        # upon submission as accruing at the date of payment.
        if paid and paymentDate:
            self.cleaned_data['accrualDate'] = paymentDate
        else:
            self.cleaned_data.pop('paymentDate',None)

        if payTo != '1' and payToUser:
            self.cleaned_data.pop('payToUser',None)
        if payTo != '2' and payToLocation:
            self.cleaned_data.pop('payToLocation',None)
        if payTo != '3' and payToName:
            self.cleaned_data.pop('payToName',None)

        if payBy == '1' and total:
            self.cleaned_data.pop('total',None)
        if payBy == '2' and hours:
            self.cleaned_data.pop('hours',None)
        return self.cleaned_data

    class Meta:
        model = ExpenseItem
        fields = ['submissionUser','payToUser','payToLocation','payToName','category','description','hours','total','reimbursement','attachment','approved','paid','paymentDate','paymentMethod','accrualDate']
        widgets = {
            'category': ExpenseCategoryWidget,
        }

    class Media:
        js = ('admin/js/admin/RelatedObjectLookups.js','jquery-ui/jquery-ui.min.js','js/expense_reporting.js')
        css = {
            'all': ('jquery-ui/jquery-ui.min.css',),
        }


class InvoiceItemChoiceField(forms.ModelChoiceField):
    '''
    This exists so that the validators for InvoiceItems (EventRegistrations) are not
    thrown off by the fact that the initial query is blank.
    '''

    def to_python(self,value):
        try:
            value = super(self.__class__,self).to_python(value)
        except (ValueError, ValidationError):
            key = self.to_field_name or 'pk'
            value = InvoiceItem.objects.filter(**{key: value})
            if not value.exists():
                raise ValidationError(self.error_messages['invalid_choice'], code='invalid_choice')
            else:
                value = value.first()
        return value


class RevenueReportingForm(forms.ModelForm):
    associateWith = forms.ChoiceField(widget=forms.RadioSelect, choices=REVENUE_ASSOCIATION_CHOICES,label=_('This revenue is associated with:'),initial=1)
    currentlyHeldBy = forms.ModelChoiceField(
        queryset=User.objects.filter(Q(is_staff=True) | Q(staffmember__isnull=False)),
        label=_('Cash currently in possession of'),
        required=False,
        widget=autocomplete.ModelSelect2(
            url='autocompleteUser',
            attrs={
                # This will set the input placeholder attribute:
                'data-placeholder': _('Enter a user name'),
                # This will set the yourlabs.Autocomplete.minimumCharacters
                # options, the naming conversion is handled by jQuery
                'data-minimum-input-length': 2,
                'data-max-results': 4,
                'class': 'modern-style',
            }
        )
    )
    paymentMethod = autocomplete.Select2ListCreateChoiceField(
        choice_list=get_method_list,
        required=False,
        widget=autocomplete.ListSelect2(url='paymentMethod-list-autocomplete'),
        label=_('Payment method'),
    )

    def __init__(self, *args, **kwargs):
        self.helper = FormHelper()
        self.helper.add_input(Submit('submit', _('Submit')))
        user = kwargs.pop('user', None)

        if hasattr(user,'id'):
            kwargs.update(initial={
                'submissionUser': user.id
            })
        super(RevenueReportingForm,self).__init__(*args,**kwargs)
        self.fields['submissionUser'].widget = forms.HiddenInput()
        self.fields['invoiceNumber'].widget = forms.HiddenInput()
        self.fields["invoiceItem"] = InvoiceItemChoiceField(queryset=InvoiceItem.objects.none(),required=False)

        # re-order fields to put the associateWith RadioSelect first.
        newFields = OrderedDict()
        newFields['associateWith'] = self.fields['associateWith']
        for key,field in self.fields.items():
            if key not in ['associateWith']:
                newFields[key] = field
        self.fields = newFields

    def clean_description(self):
        ''' Avoid empty descriptions '''
        return self.cleaned_data['description'] or _('Form Submitted Revenue')

    def clean_invoiceNumber(self):
        ''' Create a unique invoice number '''
        return 'SUBMITTED_%s_%s' % (getattr(self.cleaned_data['submissionUser'],'id','None'),timezone.now().strftime('%Y%m%d%H%M%S'))

    def clean(self):
        # Custom cleaning ensures that revenues are not attributed
        # to both a series and to an event.
        super(RevenueReportingForm, self).clean()

        associateWith = self.cleaned_data.get('associateWith')
        event = self.cleaned_data.get('event')

        if associateWith in ['1','3'] and event:
            self.cleaned_data.pop('event', None)
            self.cleaned_data.pop('invoiceItem', None)

        return self.cleaned_data

    class Meta:
        model = RevenueItem
        fields = ['submissionUser','invoiceNumber','category','description','event','invoiceItem','receivedFromName','paymentMethod','currentlyHeldBy','total','attachment']

    class Media:
        js = ('js/revenue_reporting.js',)


class CompensationRuleUpdateForm(forms.ModelForm):
    ''' Used for bulk update of Instructor compensation rules. '''

    def save(self, commit=True):
        ''' Handle the update logic for this in the view, not the form '''
        pass

    class Meta:
        model = StaffMemberWageInfo
        fields = ['category', 'rentalRate','applyRateRule','dayStarts','weekStarts','monthStarts']



from django.core.urlresolvers import reverse
from django.utils.translation import ugettext_lazy as _
from django.utils import timezone

from dateutil.relativedelta import relativedelta
from calendar import month_name

from cms.toolbar_pool import toolbar_pool
from cms.toolbar_base import CMSToolbar
from cms.toolbar.items import Break


@toolbar_pool.register
class FinancialToolbar(CMSToolbar):
    ''' Adds links to financial functions '''

    def addTheMenu(self):
        ''' Since not all staff users get this menu, this function simplifies its creation '''

        menu = self.toolbar.get_or_create_menu('financial', _('Finances'))

        # Other apps may add related items, so check if those exist yet, and if so, add before them
        startPosition = menu.find_first(
            Break,
            identifier='financial_related_items_break'
        )

        if not startPosition:
            menu.add_break('financial_related_items_break')
            startPosition = menu.find_first(
                Break,
                identifier='financial_related_items_break'
            )
        return (menu, startPosition)

    def populate(self):
        menu = None
        addBreak = False
        if self.request.user.has_perm('financial.add_expenseitem'):
            menu, newPosition = self.addTheMenu()
            menu.add_link_item(_('Submit Expenses'), url=reverse('submitExpenses'), position=newPosition)
            addBreak = True
        if self.request.user.has_perm('financial.add_revenueitem'):
            menu, newPosition = self.addTheMenu()
            menu.add_link_item(_('Submit Revenues'), url=reverse('submitRevenues'), position=newPosition)
            addBreak = True
        if hasattr(self.request.user,'staffmember') and self.request.user.staffmember and self.request.user.has_perm('core.view_own_instructor_finances'):
            menu, newPosition = self.addTheMenu()
            menu.add_link_item(_('Your Payment History'), url=reverse('instructorPayments'), position=newPosition)

        if addBreak:
            menu.add_break('post_submission_break', position=newPosition + 1)

        if self.request.user.has_perm('financial.view_finances_bymonth'):
            menu, newPosition = self.addTheMenu()
            menu.add_link_item(_('View Monthly Financial Summary'), url=reverse('financesByMonth'), position=newPosition)
        if self.request.user.has_perm('financial.view_finances_byevent'):
            menu, newPosition = self.addTheMenu()
            menu.add_link_item(_('View Financial Summary By Event'), url=reverse('financesByEvent'), position=newPosition)

        if self.request.user.has_perm('financial.view_finances_detail'):
            menu, newPosition = self.addTheMenu()
            detail_submenu = menu.get_or_create_menu('financial-details',_('Detailed Breakdown'), position=newPosition)

            now = timezone.now()
            month_ago = now - relativedelta(months=1)
            two_months_ago = now - relativedelta(months=2)
            three_months_ago = now - relativedelta(months=3)
            year_ago = now - relativedelta(years=1)

            detail_submenu.add_link_item(_('Year To Date'), url=reverse('financialDetailView', kwargs={'year': str(now.year)}))
            detail_submenu.add_link_item(_('%s' % year_ago.year), url=reverse('financialDetailView',kwargs={'year': str(year_ago.year)}))

            detail_submenu.add_break('post_annual_break')

            detail_submenu.add_link_item(_('Month To Date'), url=reverse('financialDetailView',kwargs={'year': str(now.year), 'month': month_name[now.month]}))
            detail_submenu.add_link_item(_('%s %s' % (month_name[month_ago.month], month_ago.year)), url=reverse('financialDetailView',kwargs={'year': str(month_ago.year), 'month': month_name[month_ago.month]}))
            detail_submenu.add_link_item(_('%s %s' % (month_name[two_months_ago.month], two_months_ago.year)), url=reverse('financialDetailView',kwargs={'year': str(two_months_ago.year), 'month': month_name[two_months_ago.month]}))
            detail_submenu.add_link_item(_('%s %s' % (month_name[three_months_ago.month], three_months_ago.year)), url=reverse('financialDetailView',kwargs={'year': str(three_months_ago.year), 'month': month_name[three_months_ago.month]}))

        if menu and (
            self.request.user.has_perm('financial.change_expenseitem') or self.request.user.has_perm('financial.change_revenueitem') or
            self.request.user.has_perm('financial.change_expensecategory') or self.request.user.has_perm('financial.change_revenuecategory') or
            self.request.user.has_perm('financial.change_repeatedexpenserule')
        ):
            related_menu = menu.get_or_create_menu('financial-related',_('Related Items'), position=newPosition + 2)

            if self.request.user.has_perm('financial.change_expensecategory'):
                related_menu.add_link_item(_('Expense Categories'), url=reverse('admin:financial_expensecategory_changelist'))
            if self.request.user.has_perm('financial.change_expenseitem'):
                related_menu.add_link_item(_('Expense Items'), url=reverse('admin:financial_expenseitem_changelist'))
            if self.request.user.has_perm('financial.change_repeatedexpenserule'):
                related_menu.add_link_item(_('Repeated Expense Rules'), url=reverse('admin:financial_repeatedexpenserule_changelist'))
            if self.request.user.has_perm('financial.change_revenuecategory'):
                related_menu.add_link_item(_('Revenue Categories'), url=reverse('admin:financial_revenuecategory_changelist'))
            if self.request.user.has_perm('financial.change_revenueitem'):
                related_menu.add_link_item(_('Revenue Items'), url=reverse('admin:financial_revenueitem_changelist'))



from django import template
from django.db.models.query import QuerySet
from django.core.exceptions import ObjectDoesNotExist

from calendar import month_name, day_name
from polymorphic.query import PolymorphicQuerySet
import re

from danceschool.core.models import Event, DanceRole, Instructor, Location
from danceschool.core.constants import getConstant

# This is needed to register all the tags
register = template.Library()


@register.filter
def readable_month(month):
    try:
        return month_name[month]
    except (TypeError,IndexError):
        return None


@register.filter
def urlencodename(string):
    return string.replace('-','_').replace(' ','+')


@register.filter
def urldecodename(string):
    return string.replace('_','-').replace('+',' ')


@register.simple_tag
def readable_weekday(weekday):
    try:
        return day_name[weekday]
    except (TypeError,IndexError):
        return None


@register.simple_tag
def getSchoolName():
    '''
    For easily retrieving the school name without overriding templates,
    such as in the default navbar template.
    '''
    return getConstant('contact__businessName')


@register.filter(name='has_group')
def has_group(user, group_name):
    '''
    This allows specification group-based permissions in templates.
    In most instances, creating model-based permissions and giving
    them to the desired group is preferable.
    '''
    if user.groups.filter(name=group_name).exists():
        return True
    return False


@register.filter
def get_item(dictionary, key):
    '''
    This allows us to get a specific key from a dictionary, where
    the key can be a variable name.
    '''
    return dictionary.get(key)


@register.simple_tag
def get_item_by_key(passed_list, key, value):
    '''
    This one allows us to get one or more items from a list of
    dictionaries based on the value of a specified key, where
    both the key and the value can be variable names.  Does
    not work with None or null string passed values.
    '''

    if value in [None,'']:
        return

    if type(passed_list) in [QuerySet, PolymorphicQuerySet]:
        sub_list = passed_list.filter(**{key: value})
    else:
        sub_list = [x for x in passed_list if x.get(key) == value]

    if len(sub_list) == 1:
        return sub_list[0]
    return sub_list


@register.simple_tag
def get_field_for_object(field_type,field_id, form):
    '''
    This tag allows one to get a specific series or event form field
    in registration views.
    '''
    field_name = field_type + '_' + str(field_id)
    return form.__getitem__(field_name)


@register.filter
def template_exists(template_name):
    '''
    Determine if a given template exists so that it can be loaded
    if so, or a default alternative can be used if not.
    '''
    try:
        template.loader.get_template(template_name)
        return True
    except template.TemplateDoesNotExist:
        return False


@register.simple_tag
def numRegisteredForRole(event,role):
    '''
    This tag allows one to access the number of registrations
    for any dance role.
    '''
    if not isinstance(event,Event) or not isinstance(role,DanceRole):
        return None
    return event.numRegisteredForRole(role)


@register.simple_tag
def numRegisteredForRoleName(event,roleName):
    '''
    This tag allows one to access the number of registrations
    for any dance role using only the role's name.
    '''
    if not isinstance(event,Event):
        return None

    try:
        role = DanceRole.objects.get(name=roleName)
    except ObjectDoesNotExist:
        return None

    return event.numRegisteredForRole(role)


@register.filter
def getStatusValue(obj,value):
    '''
    Several model fields use the DjangoChoices app for cleaner
    enum-type value storage of status variables.  This gets the label for a passed value
    associated with one of these fields.
    '''
    if isinstance(obj,Event):
        choice_dict = Event.RegStatus.values
    elif isinstance(obj,Instructor):
        choice_dict = Instructor.InstructorStatus.values
    elif isinstance(obj,Location):
        choice_dict = Location.StatusChoices.values
    else:
        return None

    # All these classes store choice information in the status field.
    return choice_dict.get(obj.status) or obj.status


@register.filter
def camelSpace(obj):
    ''' Add spaces in camelCase words '''
    return re.sub(r'((?<=[a-z])[A-Z]|(?<!\A)[A-Z](?=[a-z]))', r' \1', obj)



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-08-30 15:59
from __future__ import unicode_literals

import django.core.validators
from django.db import migrations, models
import django.db.models.deletion


def create_initial_category(apps, schema_editor):
    ''' Create a default category for existing discounts '''
    DiscountCategory = apps.get_model('discounts','DiscountCategory')
    db_alias = schema_editor.connection.alias
    DiscountCategory.objects.using(db_alias).create(
        id=1,name='General Discounts',order=1,cannotCombine=False
    )


def delete_initial_category(apps, schema_editor):
    ''' Category will be deleted anyway '''
    pass


class Migration(migrations.Migration):

    dependencies = [
        ('discounts', '0004_auto_20170808_2020'),
    ]

    operations = [
        migrations.CreateModel(
            name='DiscountCategory',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=30, unique=True, verbose_name='Category name')),
                ('order', models.FloatField(help_text='Discounts from categories with lower numbers are applied first.', unique=True, verbose_name='Category order')),
                ('cannotCombine', models.BooleanField(default=False, help_text='If checked, then discounts in this category cannot be combined with any other type of discount.', verbose_name='Cannot be combined')),
            ],
            options={
                'verbose_name_plural': 'Discount categories',
                'ordering': ('order',),
                'verbose_name': 'Discount category',
            },
        ),
        migrations.RenameField(
            model_name='discountcombo',
            old_name='doorGeneralPrice',
            new_name='doorPrice',
        ),
        migrations.RenameField(
            model_name='discountcombo',
            old_name='onlineGeneralPrice',
            new_name='onlinePrice',
        ),
        migrations.RemoveField(
            model_name='discountcombo',
            name='doorStudentPrice',
        ),
        migrations.RemoveField(
            model_name='discountcombo',
            name='onlineStudentPrice',
        ),
        migrations.AlterField(
            model_name='discountcombo',
            name='doorPrice',
            field=models.FloatField(blank=True, null=True, validators=[django.core.validators.MinValueValidator(0)], verbose_name='At-the-door price'),
        ),
        migrations.AlterField(
            model_name='discountcombo',
            name='onlinePrice',
            field=models.FloatField(blank=True, null=True, validators=[django.core.validators.MinValueValidator(0)], verbose_name='Online price'),
        ),
        migrations.AddField(
            model_name='discountcombo',
            name='studentsOnly',
            field=models.BooleanField(default=False, help_text='Check this box to create student discounts.', verbose_name='Discount for HS/college/university students only'),
        ),
        migrations.RunPython(create_initial_category, delete_initial_category),
        migrations.AddField(
            model_name='discountcombo',
            name='category',
            field=models.ForeignKey(default=1, help_text='One discount can be applied per category, and the order in which discounts are applied depends on the order parameter of the categories.', on_delete=django.db.models.deletion.CASCADE, to='discounts.DiscountCategory', verbose_name='Discount category'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='registrationdiscount',
            name='registration',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.Registration', verbose_name='Registration'),
        ),
        migrations.AlterField(
            model_name='temporaryregistrationdiscount',
            name='registration',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.TemporaryRegistration', verbose_name='Temporary registration'),
        ),
        migrations.AlterUniqueTogether(
            name='registrationdiscount',
            unique_together=set([('registration', 'discount')]),
        ),
        migrations.AlterUniqueTogether(
            name='temporaryregistrationdiscount',
            unique_together=set([('registration', 'discount')]),
        ),
    ]



from django.contrib.syndication.views import Feed
from django.core.urlresolvers import reverse
from django.utils.translation import ugettext_lazy as _

from .models import NewsItem

from danceschool.core.constants import getConstant


class LatestNewsFeed(Feed):

    def title(self):
        return _("%s Latest News" % getConstant('contact__businessName'))

    def description(self):
        return _("Updates and news items from %s" % getConstant('contact__businessName'))

    def link(self):
        return reverse('news_feed')

    def items(self):
        return NewsItem.objects.order_by('-publicationDate')[:10]

    def item_title(self, item):
        return item.title

    def item_pubDate(self, item):
        return item.publicationDate

    def item_description(self, item):
        return item.content

    # item_link is only needed if NewsItem has no get_absolute_url method.
    def item_link(self, item):
        return reverse('news_feed')



from django.utils.translation import ugettext_lazy as _
from django.utils import timezone

from cms.plugin_base import CMSPluginBase
from cms.plugin_pool import plugin_pool
from cms.models.pluginmodel import CMSPlugin

from datetime import datetime, timedelta

from .models import InstructorListPluginModel, LocationPluginModel, LocationListPluginModel, EventListPluginModel, Instructor, Event, Series, PublicEvent, Location
from .mixins import PluginTemplateMixin
from .registries import plugin_templates_registry, PluginTemplateBase


class InstructorListPlugin(PluginTemplateMixin, CMSPluginBase):
    model = InstructorListPluginModel
    name = _('Set of Instructor Images or Bios')
    render_template = 'core/instructor_image_set.html'
    cache = True
    module = _('Staff')

    def render(self, context, instance, placeholder):
        context = super(InstructorListPlugin,self).render(context,instance,placeholder)

        listing = Instructor.objects.all()

        if instance.statusChoices:
            listing = listing.filter(status__in=instance.statusChoices)
        else:
            listing = listing.exclude(status__in=[
                Instructor.InstructorStatus.hidden,
                Instructor.InstructorStatus.retired,
                Instructor.InstructorStatus.retiredGuest
            ])

        if instance.photoRequired:
            listing = listing.filter(image__isnull=False)
        if instance.bioRequired:
            listing = listing.filter(bio__isnull=False).exclude(bio__exact='')

        if instance.activeUpcomingOnly:
            listing = listing.filter(eventstaffmember__event__endTime__gte=timezone.now()).distinct()

        if instance.orderChoice == 'firstName':
            listing = listing.order_by('firstName','lastName')
        elif instance.orderChoice == 'status':
            listing = listing.order_by('status','lastName','firstName')
        elif instance.orderChoice == 'random':
            listing = listing.order_by('?')
        else:
            listing = listing.order_by('lastName','firstName')

        context.update({
            'list_title': instance.title,
            'instructor_list': listing,
            'thumbnail': instance.imageThumbnail,
        })

        if instance.imageThumbnail:
            context['thumbnail_dimensions'] = '%sx%s' % (instance.imageThumbnail.width, instance.imageThumbnail.height)
        return context


class LocationListPlugin(PluginTemplateMixin, CMSPluginBase):
    model = LocationListPluginModel
    name = _('Information on All Public Locations')
    render_template = 'core/location_directions.html'
    module = _('Locations')

    def render(self, context, instance, placeholder):
        ''' Allows this plugin to use templates designed for a list of locations. '''
        context = super(LocationListPlugin,self).render(context,instance,placeholder)
        context['location_list'] = Location.objects.filter(status=Location.StatusChoices.active)
        return context


class LocationPlugin(PluginTemplateMixin, CMSPluginBase):
    model = LocationPluginModel
    name = _('Individual Location Information')
    render_template = 'core/location_directions.html'
    cache = True
    module = _('Locations')

    def render(self, context, instance, placeholder):
        ''' Allows this plugin to use templates designed for a list of locations. '''
        context = super(LocationPlugin,self).render(context,instance,placeholder)
        context['location_list'] = [instance.location,]
        return context


class EventListPlugin(PluginTemplateMixin, CMSPluginBase):
    model = EventListPluginModel
    name = _('List of Events')
    cache = True
    module = _('Events')
    render_template = 'core/events_bymonth_list.html'

    fieldsets = (
        (None, {
            'fields': ('title','eventType','template','cssClasses')
        }),
        (_('Limit Start Date'), {
            'fields': ('limitTypeStart','daysStart','startDate'),
        }),
        (_('Limit End Date'), {
            'fields': ('limitTypeEnd','daysEnd','endDate'),
        }),
        (_('Other Restrictions'), {
            'fields': ('limitToOpenRegistration','location','weekday'),
        })
    )

    def render(self, context, instance, placeholder):
        context = super(EventListPlugin,self).render(context,instance,placeholder)

        listing = Event.objects.exclude(status__in=[Event.RegStatus.hidden, Event.RegStatus.linkOnly])

        if instance.eventType == 'S':
            listing = listing.instance_of(Series)
        elif instance.eventType == 'P':
            listing = listing.instance_of(PublicEvent)

        filters = {}

        startKey = 'endTime__gte'
        endKey = 'startTime__lte'

        if instance.limitTypeStart == 'S':
            startKey = 'startTime__gte'
        if instance.limitTypeEnd == 'E':
            endKey = 'endTime__lte'

        if instance.startDate:
            filters[startKey] = datetime.combine(instance.startDate,datetime.min.time())
        elif instance.daysStart is not None:
            filters[startKey] = timezone.now() + timedelta(days=instance.daysStart)

        if instance.endDate:
            filters[endKey] = datetime.combine(instance.endDate,datetime.max.time())
        elif instance.daysEnd is not None:
            filters[endKey] = timezone.now() + timedelta(days=instance.daysEnd)

        if instance.limitToOpenRegistration:
            filters['registrationOpen'] = True

        if instance.location:
            filters['location'] = instance.location

        # Python calendar module indexes weekday differently from Django
        if instance.weekday is not None:
            filters['startTime__week_day'] = (instance.weekday + 2) % 7

        listing = listing.filter(**filters)

        context.update({
            'event_list': listing,
        })
        return context


class PublicCalendarPlugin(CMSPluginBase):
    model = CMSPlugin
    name = _('Public Calendar')
    render_template = 'core/public_calendar.html'
    cache = True
    module = _('Events')


plugin_pool.register_plugin(InstructorListPlugin)
plugin_pool.register_plugin(LocationPlugin)
plugin_pool.register_plugin(LocationListPlugin)
plugin_pool.register_plugin(EventListPlugin)
plugin_pool.register_plugin(PublicCalendarPlugin)


@plugin_templates_registry.register
class InstructorImageSetTemplate(PluginTemplateBase):
    plugin = 'InstructorListPlugin'
    template_name = 'core/instructor_image_set.html'
    description = _('Instructor Images')


@plugin_templates_registry.register
class InstructorListTemplate(PluginTemplateBase):
    plugin = 'InstructorListPlugin'
    template_name = 'core/instructor_list.html'
    description = _('Instructor Bios with Images')


@plugin_templates_registry.register
class LocationDirectionsTemplate(PluginTemplateBase):
    plugin = 'LocationListPlugin'
    template_name = 'core/location_directions.html'
    description = _('Directions for Each Location')


@plugin_templates_registry.register
class SingleLocationDirectionsTemplate(PluginTemplateBase):
    plugin = 'LocationPlugin'
    template_name = 'core/location_directions.html'
    description = _('Directions for This Location')


@plugin_templates_registry.register
class EventListPluginTemplate(PluginTemplateBase):
    plugin = 'EventListPlugin'
    template_name = 'core/events_bymonth_list.html'
    description = _('Default List of Events By Month')



from django.views.generic import DetailView, TemplateView, CreateView, View, FormView
from django.shortcuts import get_object_or_404
from django.http import HttpResponse, Http404, HttpResponseBadRequest
from django.db.models import Q, Sum, F, Min
from django.contrib.auth.models import User
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist
from django.utils.translation import ugettext_lazy as _
from django.utils import timezone
from django.contrib.messages.views import SuccessMessageMixin

from datetime import datetime
import unicodecsv as csv
from calendar import month_name
from urllib.parse import unquote_plus
from braces.views import PermissionRequiredMixin, StaffuserRequiredMixin, UserFormKwargsMixin

from danceschool.core.models import Instructor, Location, Event, StaffMember
from danceschool.core.constants import getConstant
from danceschool.core.mixins import StaffMemberObjectMixin, FinancialContextMixin, AdminSuccessURLMixin
from danceschool.core.utils.timezone import ensure_timezone
from danceschool.core.utils.requests import getIntFromGet, getDateTimeFromGet

from .models import ExpenseItem, RevenueItem, ExpenseCategory, RevenueCategory, RepeatedExpenseRule
from .helpers import prepareFinancialStatement, getExpenseItemsCSV, getRevenueItemsCSV, prepareStatementByMonth, prepareStatementByEvent
from .forms import ExpenseReportingForm, RevenueReportingForm, CompensationRuleUpdateForm
from .constants import EXPENSE_BASES


class ExpenseReportingView(AdminSuccessURLMixin, StaffuserRequiredMixin, UserFormKwargsMixin, SuccessMessageMixin, CreateView):
    template_name = 'cms/forms/display_crispy_form_classbased_admin.html'
    form_class = ExpenseReportingForm
    success_message = _('Expense item successfully submitted.')

    def get_context_data(self,**kwargs):
        context = super(ExpenseReportingView,self).get_context_data(**kwargs)

        context.update({
            'form_title': _('Report Expenses'),
            'form_description': _('Use this form to report expenses.'),
        })
        return context


class RevenueReportingView(AdminSuccessURLMixin, StaffuserRequiredMixin, UserFormKwargsMixin, SuccessMessageMixin, CreateView):
    template_name = 'cms/forms/display_crispy_form_classbased_admin.html'
    form_class = RevenueReportingForm
    success_message = _('Revenue item successfully submitted.')

    def get_context_data(self,**kwargs):
        context = super(RevenueReportingView,self).get_context_data(**kwargs)

        context.update({
            'form_title': _('Report Revenues'),
            'form_description': _('Use this form to report revenues.'),
        })
        return context


class InstructorPaymentsView(StaffMemberObjectMixin, PermissionRequiredMixin, DetailView):
    model = Instructor
    template_name = 'financial/instructor_payments.html'
    permission_required = 'core.view_own_instructor_finances'
    as_csv = False

    def get_context_data(self,**kwargs):
        instructor = self.object
        context = {}

        query_filter = Q()

        # These will be passed to the template
        year = self.kwargs.get('year')
        eligible_years = list(set([x.year for x in ExpenseItem.objects.values_list('accrualDate',flat=True).distinct()]))
        eligible_years.sort(reverse=True)

        if not year or year == 'all':
            int_year = None
            year = 'all'
        else:
            try:
                int_year = int(year)

                # Check for year in kwargs and ensure that it is eligible
                if int_year not in eligible_years:
                    raise Http404(_("Invalid year."))
                query_filter = query_filter & (Q(accrualDate__year=int_year) | Q(paymentDate__year=int_year) | Q(submissionDate__year=int_year))
            except (ValueError, TypeError):
                raise Http404(_("Invalid year."))

        # No point in continuing if we can't actually match this instructor to their payments.
        if not hasattr(instructor,'userAccount'):
            return super(DetailView, self).get_context_data(instructor=instructor)

        all_payments = instructor.userAccount.payToUser.filter(query_filter).order_by('-submissionDate')

        paid_items = all_payments.filter(**{'paid':True,'reimbursement':False}).order_by('-paymentDate')
        unpaid_items = all_payments.filter(**{'paid':False}).order_by('-submissionDate')
        reimbursement_items = all_payments.filter(**{'paid':True,'reimbursement':True}).order_by('-paymentDate')

        if int_year:
            time_lb = ensure_timezone(datetime(int_year,1,1,0,0))
            time_ub = ensure_timezone(datetime(int_year + 1,1,1,0,0))
        else:
            time_lb = ensure_timezone(datetime(timezone.now().year,1,1,0,0))
            time_ub = ensure_timezone(datetime(timezone.now().year + 1,1,1,0,0))

        paid_this_year = paid_items.filter(paymentDate__gte=time_lb,paymentDate__lt=time_ub).order_by('-paymentDate')
        accrued_paid_this_year = paid_items.filter(accrualDate__gte=time_lb,accrualDate__lt=time_ub).order_by('-paymentDate')
        reimbursements_this_year = all_payments.filter(paymentDate__gte=time_lb,paymentDate__lt=time_ub,paid=True,reimbursement=True)

        context.update({
            'instructor': instructor,
            'current_year': year,
            'eligible_years': eligible_years,
            'all_payments': all_payments,
            'paid_items': paid_items,
            'unpaid_items': unpaid_items,
            'reimbursement_items': reimbursement_items,
            'paid_this_year': paid_this_year,
            'accrued_paid_this_year': accrued_paid_this_year,
            'reimbursements_this_year': reimbursements_this_year,
            'total_paid_alltime': sum(filter(None,[x.total for x in paid_items])),
            'total_awaiting_payment': sum(filter(None,[x.total for x in unpaid_items])),
            'total_paid_this_year': sum(filter(None,[x.total for x in paid_this_year])),
            'total_reimbursements': sum(filter(None,[x.total for x in reimbursements_this_year])),
        })

        # Note: This get the detailview's context, not all the mixins.  Supering itself led to an infinite loop.
        return super(DetailView, self).get_context_data(**context)

    def dispatch(self, request, *args, **kwargs):
        if 'as_csv' in kwargs:
            self.as_csv = True
        return super(InstructorPaymentsView, self).dispatch(request, *args, **kwargs)

    def render_to_response(self, context, **response_kwargs):
        if self.as_csv:
            return self.render_to_csv(context)
        return super(InstructorPaymentsView, self).render_to_response(context, **response_kwargs)

    def render_to_csv(self, context):
        instructor = context['instructor']
        if hasattr(instructor.userAccount,'payToUser'):
            all_expenses = context['all_payments']
        else:
            all_expenses = ExpenseItem.objects.none()
        return getExpenseItemsCSV(all_expenses,scope='instructor')


class OtherInstructorPaymentsView(InstructorPaymentsView):
    permission_required = 'core.view_other_instructor_finances'

    def get_object(self, queryset=None):
        if 'first_name' in self.kwargs and 'last_name' in self.kwargs:
            return get_object_or_404(
                Instructor.objects.filter(**{'firstName': unquote_plus(self.kwargs['first_name']).replace('_',' '), 'lastName': unquote_plus(self.kwargs['last_name']).replace('_',' ')})
            )
        else:
            return None


class FinancesByEventView(PermissionRequiredMixin, TemplateView):
    permission_required = 'financial.view_finances_byevent'
    cache_timeout = 3600
    template_name = 'financial/finances_byevent.html'
    as_csv = False
    paginate_by = 25

    def get_paginate_by(self,queryset=None):
        if self.as_csv:
            return 1000
        else:
            return self.paginate_by

    def get_context_data(self,**kwargs):
        context = {}

        # These will be passed to the template
        year = self.kwargs.get('year')
        eligible_years = list(set([x.year for x in ExpenseItem.objects.values_list('accrualDate',flat=True).distinct()]))
        eligible_years.sort(reverse=True)

        if not year or year == 'all':
            int_year = None
            year = 'all'
        else:
            try:
                int_year = int(year)

                # Check for year in kwargs and ensure that it is eligible
                if int_year not in eligible_years:
                    raise Http404(_("Invalid year."))
            except (ValueError, TypeError):
                raise Http404(_("Invalid year."))

        context['current_year'] = year
        context['eligible_years'] = eligible_years

        page = self.kwargs.get('page') or self.request.GET.get('page') or 1

        context['statement'] = prepareFinancialStatement(year=int_year)
        paginator, page_obj, statementByEvent, is_paginated = prepareStatementByEvent(year=int_year,page=page,paginate_by=self.get_paginate_by())
        context.update({
            'paginator': paginator,
            'page_obj': page_obj,
            'is_paginated': is_paginated,
        })
        context['statement']['statementByEvent'] = statementByEvent

        # Get a list of all roles with positive registrations in the statement:
        role_set = set()
        for x in statementByEvent:
            role_set.update(list(x.get('registrations').keys()))
        role_list = list(role_set)
        sorted(role_list, key=lambda x: (x is None, x))
        context['roles'] = role_list

        return super(self.__class__,self).get_context_data(**context)

    def dispatch(self, request, *args, **kwargs):
        if 'as_csv' in kwargs:
            self.as_csv = True
        return super(self.__class__, self).dispatch(request, *args, **kwargs)

    def render_to_response(self, context, **response_kwargs):
        if self.as_csv:
            return self.render_to_csv(context)
        return super(self.__class__, self).render_to_response(context, **response_kwargs)

    def render_to_csv(self, context):
        statement = context['statement']
        response = HttpResponse(content_type='text/csv')
        response['Content-Disposition'] = 'attachment; filename="financialStatementByEvent.csv"'

        writer = csv.writer(response, csv.excel)
        response.write(u'\ufeff'.encode('utf8'))  # BOM (optional...Excel needs it to open UTF-8 file properly)

        header_list = [
            _('Event'),
            _('Month'),
            _('Registrations: Total'),
            _('Registrations: Leads'),
            _('Registrations: Follows'),
            _('Revenues: Gross'),
            _('Revenues: Net'),
            _('Expenses: Instruction'),
            _('Expenses: Venue'),
            _('Expenses: Other'),
            _('Expenses: Total'),
            _('Net Profit'),
        ]
        writer.writerow(header_list)

        for x in statement['statementByEvent']:
            this_row_data = [
                x['event_name'],
                x['month_name'],
                x['registrations']['total'],
                x['registrations']['leads'],
                x['registrations']['follows'],
                x['revenues']['gross'],
                x['revenues']['net'],
                x['expenses']['instruction'],
                x['expenses']['venue'],
                x['expenses']['other'],
                x['expenses']['total'],
                x['net_profit'],
            ]
            writer.writerow(this_row_data)

        return response


class FinancesByMonthView(PermissionRequiredMixin, TemplateView):
    permission_required = 'financial.view_finances_bymonth'
    cache_timeout = 3600
    template_name = 'financial/finances_bymonth.html'
    as_csv = False
    paginate_by = 24

    def get_paginate_by(self,queryset=None):
        if self.as_csv:
            return 1000
        else:
            return self.paginate_by

    def get(self,request,*args,**kwargs):
        '''
        Allow passing of basis and time limitations
        '''
        try:
            year = int(self.kwargs.get('year'))
        except (ValueError, TypeError):
            year = getIntFromGet(request,'year')

        kwargs.update({
            'year': year,
            'basis': request.GET.get('basis'),
        })

        if kwargs.get('basis') not in EXPENSE_BASES.keys():
            kwargs['basis'] = 'accrualDate'

        return super(self.__class__,self).get(request,*args,**kwargs)

    def get_context_data(self,**kwargs):
        context = {}

        # Determine the period over which the statement should be produced.
        year = kwargs.get('year')

        eligible_years = list(set([x.year for x in ExpenseItem.objects.values_list('accrualDate',flat=True).distinct()]))
        eligible_years.sort(reverse=True)

        if year and year not in eligible_years:
            raise Http404(_("Invalid year."))

        context.update({
            'basis': kwargs.get('basis'),
            'basis_name': EXPENSE_BASES[kwargs.get('basis')],
            'year': year,
            'current_year': year or 'all',
            'eligible_years': eligible_years,
        })

        page = self.kwargs.get('page') or self.request.GET.get('page') or 1

        context['statement'] = prepareFinancialStatement(year=year)
        paginator, page_obj, statementByMonth, is_paginated = prepareStatementByMonth(year=year,basis=context['basis'],page=page,paginate_by=self.get_paginate_by())
        context.update({
            'paginator': paginator,
            'page_obj': page_obj,
            'is_paginated': is_paginated,
        })
        context['statement']['statementByMonth'] = statementByMonth

        return super(self.__class__,self).get_context_data(**context)

    def dispatch(self, request, *args, **kwargs):
        if 'as_csv' in kwargs:
            self.as_csv = True
        return super(self.__class__, self).dispatch(request, *args, **kwargs)

    def render_to_response(self, context, **response_kwargs):
        if self.as_csv:
            return self.render_to_csv(context)
        return super(self.__class__, self).render_to_response(context, **response_kwargs)

    def render_to_csv(self, context):
        statement = context['statement']
        response = HttpResponse(content_type='text/csv')
        response['Content-Disposition'] = 'attachment; filename="financialStatementByMonth.csv"'

        writer = csv.writer(response, csv.excel)
        response.write(u'\ufeff'.encode('utf8'))  # BOM (optional...Excel needs it to open UTF-8 file properly)

        header_list = [
            _('Month Name'),
            _('Revenues: Net'),
            _('Expenses: Instruction'),
            _('Expenses: Venue'),
            _('Expenses: Other'),
            _('Expenses: Total'),
            _('Registrations'),
            _('Net Profit'),
        ]
        writer.writerow(header_list)

        for x in statement['statementByMonth']:
            this_row_data = [
                x['month_name'],
                x['revenues'],
                x['expenses']['instruction'],
                x['expenses']['venue'],
                x['expenses']['other'],
                x['expenses']['total'],
                x['registrations'],
                x['net_profit'],
            ]
            writer.writerow(this_row_data)

        return response


class FinancialDetailView(FinancialContextMixin, PermissionRequiredMixin, TemplateView):
    permission_required = 'financial.view_finances_detail'
    template_name = 'financial/finances_detail.html'

    def get(self,request,*args,**kwargs):
        '''
        Pass any permissable GET data.  URL parameters override GET parameters
        '''
        try:
            year = int(self.kwargs.get('year'))
        except (ValueError, TypeError):
            year = getIntFromGet(request,'year')

        if self.kwargs.get('month'):
            try:
                month = int(self.kwargs.get('month'))
            except (ValueError, TypeError):
                try:
                    month = list(month_name).index(self.kwargs.get('month').title())
                except (ValueError, TypeError):
                    month = None
        else:
            month = getIntFromGet(request,'month')

        kwargs.update({
            'year': year,
            'month': month,
            'startDate': getDateTimeFromGet(request,'startDate'),
            'endDate': getDateTimeFromGet(request,'endDate'),
            'basis': request.GET.get('basis'),
        })

        if kwargs.get('basis') not in EXPENSE_BASES.keys():
            kwargs['basis'] = 'accrualDate'

        context = self.get_context_data(**kwargs)
        return self.render_to_response(context)

    def get_context_data(self,**kwargs):
        context = kwargs.copy()
        timeFilters = {}

        # Determine the period over which the statement should be produced.
        year = kwargs.get('year')
        month = kwargs.get('month')
        startDate = kwargs.get('startDate')
        endDate = kwargs.get('endDate')

        basis = kwargs.get('basis')

        context.update({
            'basis': basis,
            'basis_name': EXPENSE_BASES[basis],
            'rangeTitle': '',
        })

        if startDate:
            timeFilters['%s__gte' % basis] = startDate
            context['rangeType'] = 'Date Range'
            context['rangeTitle'] += _('From %s' % startDate.strftime('%b. %d, %Y'))
        if endDate:
            timeFilters['%s__lt' % basis] = endDate
            context['rangeType'] = 'Date Range'
            context['rangeTitle'] += _('To %s' % endDate.strftime('%b. %d, %Y'))

        if not startDate and not endDate:
            if month and year:
                end_month = ((month) % 12) + 1
                end_year = year
                if end_month == 1:
                    end_year = year + 1

                timeFilters['%s__gte' % basis] = ensure_timezone(datetime(year,month,1))
                timeFilters['%s__lt' % basis] = ensure_timezone(datetime(end_year,end_month,1))

                context['rangeType'] = 'Month'
                context['rangeTitle'] = '%s %s' % (month_name[month], year)

            elif year:
                timeFilters['%s__gte' % basis] = ensure_timezone(datetime(year,1,1))
                timeFilters['%s__lt' % basis] = ensure_timezone(datetime(year + 1,1,1))

                context['rangeType'] = 'Year'
                context['rangeTitle'] = '%s' % year
            else:
                # Assume year to date if nothing otherwise specified
                timeFilters['%s__gte' % basis] = ensure_timezone(datetime(timezone.now().year,1,1))
                timeFilters['%s__lt' % basis] = ensure_timezone(datetime(timezone.now().year + 1,1,1))

                context['rangeType'] = 'YTD'
                context['rangeTitle'] = _('Calendar Year To Date')

        context['startDate'] = timeFilters['%s__gte' % basis]
        context['endDate'] = timeFilters['%s__lt' % basis]

        # Revenues are booked on receipt basis, not payment/approval basis
        rev_timeFilters = timeFilters.copy()
        rev_basis = basis

        if basis in ['paymentDate', 'approvalDate']:
            rev_basis = 'receivedDate'
            rev_timeFilters['receivedDate__gte'] = rev_timeFilters['%s__gte' % basis]
            rev_timeFilters['receivedDate__lt'] = rev_timeFilters['%s__lt' % basis]
            del rev_timeFilters['%s__gte' % basis]
            del rev_timeFilters['%s__lt' % basis]

        expenseItems = ExpenseItem.objects.filter(**timeFilters).annotate(net=F('total') + F('adjustments') + F('fees'),basisDate=Min(basis)).order_by(basis)
        revenueItems = RevenueItem.objects.filter(**rev_timeFilters).annotate(net=F('total') + F('adjustments') - F('fees'),basisDate=Min(rev_basis)).order_by(rev_basis)

        context['expenseItems'] = expenseItems
        context['revenueItems'] = revenueItems

        # Registration revenues, instruction and venue expenses
        # are broken out separately.

        context.update({
            'instructionExpenseItems': expenseItems.filter(category__in=[getConstant('financial__classInstructionExpenseCat'),getConstant('financial__assistantClassInstructionExpenseCat')]).order_by('payToUser__last_name','payToUser__first_name'),
            'venueExpenseItems': expenseItems.filter(category=getConstant('financial__venueRentalExpenseCat')).order_by('payToLocation__name'),
            'otherExpenseItems': expenseItems.exclude(category__in=[getConstant('financial__classInstructionExpenseCat'),getConstant('financial__assistantClassInstructionExpenseCat'),getConstant('financial__venueRentalExpenseCat')]).order_by('category'),
            'expenseCategoryTotals': ExpenseCategory.objects.filter(expenseitem__in=expenseItems).annotate(category_total=Sum('expenseitem__total'),category_adjustments=Sum('expenseitem__adjustments'),category_fees=Sum('expenseitem__fees')).annotate(category_net=F('category_total') + F('category_adjustments') + F('category_fees')),
        })
        context.update({
            'instructionExpenseInstructorTotals': User.objects.filter(payToUser__in=context['instructionExpenseItems']).annotate(instructor_total=Sum('payToUser__total'),instructor_adjustments=Sum('payToUser__adjustments'),instructor_fees=Sum('payToUser__fees')).annotate(instructor_net=F('instructor_total') + F('instructor_adjustments') + F('instructor_fees')),
            'instructionExpenseOtherTotal': context['instructionExpenseItems'].filter(payToUser__isnull=True).annotate(net=F('total') + F('adjustments') + F('fees')).aggregate(instructor_total=Sum('total'),instructor_adjustments=Sum('adjustments'),instructor_fees=Sum('fees'),instructor_net=Sum('net')),

            'venueExpenseVenueTotals': Location.objects.filter(expenseitem__in=context['venueExpenseItems']).annotate(location_total=Sum('expenseitem__total'),location_adjustments=Sum('expenseitem__adjustments'),location_fees=Sum('expenseitem__fees')).annotate(location_net=F('location_total') + F('location_adjustments') + F('location_fees')),
            'venueExpenseOtherTotal': context['venueExpenseItems'].filter(payToLocation__isnull=True).annotate(location_net=F('total') + F('adjustments') + F('fees')).aggregate(location_total=Sum('total'),location_adjustments=Sum('adjustments'),location_fees=Sum('fees'),location_net=Sum('net')),

            'totalInstructionExpenses': sum([x.category_net or 0 for x in context['expenseCategoryTotals'].filter(id__in=[getConstant('financial__classInstructionExpenseCat').id,getConstant('financial__assistantClassInstructionExpenseCat').id])]),
            'totalVenueExpenses': sum([x.category_net or 0 for x in context['expenseCategoryTotals'].filter(id=getConstant('financial__venueRentalExpenseCat').id)]),
            'totalOtherExpenses': sum([x.category_net or 0 for x in context['expenseCategoryTotals'].exclude(id__in=[getConstant('financial__classInstructionExpenseCat').id,getConstant('financial__assistantClassInstructionExpenseCat').id,getConstant('financial__venueRentalExpenseCat').id])]),

            'totalExpenses': sum([x.category_net or 0 for x in context['expenseCategoryTotals']]),
        })

        context.update({
            'registrationRevenueItems': revenueItems.filter(category=getConstant('financial__registrationsRevenueCat')).order_by('-event__startTime','event__uuid'),
            'otherRevenueItems': revenueItems.exclude(category=getConstant('financial__registrationsRevenueCat')).order_by('category'),
            'revenueCategoryTotals': RevenueCategory.objects.filter(revenueitem__in=revenueItems).annotate(category_total=Sum('revenueitem__total'),category_adjustments=Sum('revenueitem__adjustments'),category_fees=Sum('revenueitem__fees')).annotate(category_net=F('category_total') + F('category_adjustments') - F('category_fees')),
        })
        context.update({
            'registrationRevenueEventTotals': Event.objects.filter(eventregistration__invoiceitem__revenueitem__in=context['registrationRevenueItems']).annotate(event_total=Sum('eventregistration__invoiceitem__revenueitem__total'),event_adjustments=Sum('eventregistration__invoiceitem__revenueitem__adjustments'),event_fees=Sum('eventregistration__invoiceitem__revenueitem__fees')).annotate(event_net=F('event_total') + F('event_adjustments') - F('event_fees')),
            'registrationRevenueOtherTotal': context['registrationRevenueItems'].filter(invoiceItem__finalEventRegistration__isnull=True).annotate(event_net=F('total') + F('adjustments') - F('fees')).aggregate(event_total=Sum('total'),event_adjustments=Sum('adjustments'),event_fees=Sum('fees'),event_net=Sum('net')),

            'totalRegistrationRevenues': sum([x.category_net or 0 for x in context['revenueCategoryTotals'].filter(id=getConstant('financial__registrationsRevenueCat').id)]),
            'totalOtherRevenues': sum([x.category_net or 0 for x in context['revenueCategoryTotals'].exclude(id=getConstant('financial__registrationsRevenueCat').id)]),
            'totalRevenues': sum([x.category_net or 0 for x in context['revenueCategoryTotals']]),
        })

        context.update({
            'netProfit': context['totalRevenues'] - context['totalExpenses'],
        })

        return super(self.__class__,self).get_context_data(**context)


class CompensationRuleUpdateView(SuccessMessageMixin, AdminSuccessURLMixin, PermissionRequiredMixin, FinancialContextMixin, FormView):
    '''
    This view is for an admin action to repeat events.
    '''
    template_name = 'financial/update_staff_compensation_rules.html'
    form_class = CompensationRuleUpdateForm
    permission_required = 'core.change_staffmember'
    success_message = _('Staff member compensation rules updated successfully.')

    def dispatch(self, request, *args, **kwargs):
        ids = request.GET.get('ids')
        ct = getIntFromGet(request,'ct')

        try:
            contentType = ContentType.objects.get(id=ct)
            self.objectClass = contentType.model_class()
        except (ValueError, ObjectDoesNotExist):
            return HttpResponseBadRequest(_('Invalid content type passed.'))

        # This view only deals with subclasses of StaffMember (Instructor, etc.)
        if not isinstance(self.objectClass(),StaffMember):
            return HttpResponseBadRequest(_('Invalid content type passed.'))

        try:
            self.queryset = self.objectClass.objects.filter(id__in=[int(x) for x in ids.split(',')])
        except ValueError:
            return HttpResponseBadRequest(_('Invalid ids passed'))

        return super(CompensationRuleUpdateView,self).dispatch(request,*args,**kwargs)

    def get_context_data(self,**kwargs):
        context = super(CompensationRuleUpdateView,self).get_context_data(**kwargs)
        context.update({
            'staffmembers': self.queryset,
            'rateRuleValues': RepeatedExpenseRule.RateRuleChoices.values,
        })

        return context

    def form_valid(self, form):
        category = form.cleaned_data.pop('category',None)

        for staffmember in self.queryset:
            staffmember.expenserules.update_or_create(
                category=category,
                defaults=form.cleaned_data,
            )

        return super(CompensationRuleUpdateView,self).form_valid(form)


class AllExpensesViewCSV(PermissionRequiredMixin, View):
    permission_required = 'financial.export_financial_data'

    def dispatch(self, request, *args, **kwargs):
        all_expenses = ExpenseItem.objects.order_by('-paid','-approved','-submissionDate')
        return getExpenseItemsCSV(all_expenses,scope='all')


class AllRevenuesViewCSV(PermissionRequiredMixin, View):
    permission_required = 'financial.export_financial_data'

    def dispatch(self, request, *args, **kwargs):
        all_revenues = RevenueItem.objects.order_by('-submissionDate')
        return getRevenueItemsCSV(all_revenues)



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-08-25 10:42
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import filer.fields.image
import jsonfield.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('filer', '0007_auto_20161016_1055'),
    ]

    operations = [
        migrations.CreateModel(
            name='BanFlaggedRecord',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('dateTime', models.DateTimeField(auto_now_add=True, verbose_name='Date and time')),
                ('ipAddress', models.GenericIPAddressField(blank=True, null=True, verbose_name='IP address')),
                ('flagCode', models.CharField(help_text='Search for this code for easier reference.', max_length=8, verbose_name='Flag code')),
                ('data', jsonfield.fields.JSONField(default={}, verbose_name='Session and form data')),
            ],
            options={
                'verbose_name_plural': 'Records of bans being flagged',
                'verbose_name': 'Record of ban being flagged',
                'ordering': ('-dateTime',),
            },
        ),
        migrations.CreateModel(
            name='BannedEmail',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('email', models.EmailField(max_length=254, unique=True, verbose_name='Email address')),
            ],
            options={
                'verbose_name_plural': 'Banned email addresses',
                'verbose_name': 'Banned email address',
            },
        ),
        migrations.CreateModel(
            name='BannedPerson',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('firstName', models.CharField(max_length=30, verbose_name='First name')),
                ('lastName', models.CharField(max_length=30, verbose_name='First name')),
                ('notes', models.TextField(blank=True, null=True, verbose_name='Notes for internal purposes')),
                ('expirationDate', models.DateTimeField(blank=True, help_text='Leave blank for no expiration.', null=True, verbose_name='Expiration date')),
                ('disabled', models.BooleanField(default=False, help_text='<strong>Note:</strong>Checking this box will prevent this ban from being automatically enforced.', verbose_name='Disabled')),
                ('submissionDate', models.DateTimeField(auto_now_add=True, verbose_name='Submission date')),
                ('modifiedDate', models.DateTimeField(auto_now=True, verbose_name='Last updated')),
                ('photo', filer.fields.image.FilerImageField(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='banned_person_image', to='filer.Image', verbose_name='Photo')),
            ],
            options={
                'permissions': (('view_banlist', 'Can view the list of banned individuals.'), ('ignore_ban', 'Can register users despite banned credentials')),
                'verbose_name_plural': 'Banned individuals',
                'verbose_name': 'Banned individual',
                'ordering': ('lastName', 'firstName'),
            },
        ),
        migrations.AddField(
            model_name='bannedemail',
            name='person',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='banlist.BannedPerson', verbose_name='Individual'),
        ),
        migrations.AddField(
            model_name='banflaggedrecord',
            name='person',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='banlist.BannedPerson', verbose_name='Person'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-06-20 21:46
from __future__ import unicode_literals

import colorful.fields
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('private_events', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='privateeventcategory',
            name='displayColor',
            field=colorful.fields.RGBColorField(default='#0000FF', help_text='Choose a color for the calendar display.'),
        ),
    ]



from django.forms import ModelForm, ModelChoiceField
from django.contrib import admin
from django.utils.translation import ugettext_lazy as _

from dal import autocomplete

from .models import DiscountCategory, DiscountCombo, DiscountComboComponent, PointGroup, PricingTierGroup, RegistrationDiscount, TemporaryRegistrationDiscount, CustomerDiscount
from danceschool.core.models import Registration, TemporaryRegistration, PricingTier, Customer


class DiscountCategoryAdmin(admin.ModelAdmin):
    list_display = ('name','order','cannotCombine')
    list_editable = ('order',)
    list_filter = ('cannotCombine',)
    search_fields = ('name',)


class DiscountComboComponentInline(admin.StackedInline):
    model = DiscountComboComponent
    extra = 1
    fields = (('pointGroup','quantity',),'allWithinPointGroup',('level','weekday'),)


class CustomerDiscountInlineForm(ModelForm):
    customer = ModelChoiceField(
        queryset=Customer.objects.all(),
        widget=autocomplete.ModelSelect2(
            url='autocompleteCustomer',
            attrs={
                # This will set the input placeholder attribute:
                'data-placeholder': _('Enter a customer name'),
                # This will set the yourlabs.Autocomplete.minimumCharacters
                # options, the naming conversion is handled by jQuery
                'data-minimum-input-length': 2,
                'data-max-results': 4,
                'class': 'modern-style',
            }
        )
    )

    class Meta:
        model = CustomerDiscount
        exclude = []


class CustomerDiscountInline(admin.StackedInline):
    model = CustomerDiscount
    form = CustomerDiscountInlineForm
    extra = 1
    classes = ['collapse',]


class DiscountComboAdminForm(ModelForm):

    class Meta:
        model = DiscountCombo
        exclude = []

    class Media:
        js = ('js/discountcombo_collapsetypes.js',)


class DiscountComboAdmin(admin.ModelAdmin):
    inlines = [DiscountComboComponentInline,CustomerDiscountInline]
    form = DiscountComboAdminForm

    list_display = ('name','category','discountType','active','expirationDate','restrictions')
    list_filter = ('category','discountType','active','newCustomersOnly','expirationDate')
    ordering = ('name',)
    actions = ['enableDiscount','disableDiscount']

    fieldsets = (
        (None, {
            'fields': ('name','category',('active','expirationDate'),'newCustomersOnly','studentsOnly','daysInAdvanceRequired','discountType',)
        }),
        (_('Flat-Price Discount (in default currency)'), {
            'classes': ('type_flatPrice',),
            'fields': ('onlinePrice','doorPrice'),
        }),
        (_('Dollar Discount (in default currency)'), {
            'classes': ('type_dollarDiscount',),
            'fields': ('dollarDiscount',),
        }),
        (_('Percentage Discount'), {
            'classes': ('type_percentageDiscount',),
            'fields': ('percentDiscount','percentUniversallyApplied'),
        }),
    )

    def restrictions(self,obj):
        text = []
        if obj.studentsOnly:
            text.append(_('Students only'))
        if obj.newCustomersOnly:
            text.append(_('First-time customer'))
        if obj.daysInAdvanceRequired:
            text.append(_('%s day advance registration' % obj.daysInAdvanceRequired))
        return ', '.join([str(x) for x in text])
    restrictions.short_description = _('Restrictions')

    def disableDiscount(self, request, queryset):
        rows_updated = queryset.update(active=False)
        if rows_updated == 1:
            message_bit = "1 discount was"
        else:
            message_bit = "%s discounts were" % rows_updated
        self.message_user(request, "%s successfully disabled." % message_bit)
    disableDiscount.short_description = _('Disable selected Discounts')

    def enableDiscount(self, request, queryset):
        rows_updated = queryset.update(active=True)
        if rows_updated == 1:
            message_bit = "1 discount was"
        else:
            message_bit = "%s discounts were" % rows_updated
        self.message_user(request, "%s successfully enabled." % message_bit)
    enableDiscount.short_description = _('Enable selected Discounts')


class RegistrationDiscountInline(admin.TabularInline):
    model = RegistrationDiscount
    readonly_fields = ('discount', 'discountAmount')
    extra = 0

    # Prevents adding new discounts without going through
    # the standard registration process
    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


class TemporaryRegistrationDiscountInline(admin.TabularInline):
    model = TemporaryRegistrationDiscount
    readonly_fields = ('discount', 'discountAmount')
    extra = 0

    # Prevents adding new discounts without going through
    # the standard registration process
    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


class PricingTierGroupInline(admin.TabularInline):
    model = PricingTierGroup
    extra = 0
    verbose_name = _('pricing tier discount group')
    verbose_name_plural = _('pricing tier discount groups')


class PointGroupAdmin(admin.ModelAdmin):
    inlines = (PricingTierGroupInline,)

    list_display = ('name',)
    ordering = ('name',)


# This adds the inlines to Registration and TemporyRegistration without subclassing
admin.site._registry[Registration].inlines.insert(0,RegistrationDiscountInline)
admin.site._registry[TemporaryRegistration].inlines.insert(0,TemporaryRegistrationDiscountInline)
admin.site._registry[PricingTier].inlines.insert(0,PricingTierGroupInline)

admin.site.register(DiscountCategory, DiscountCategoryAdmin)
admin.site.register(DiscountCombo,DiscountComboAdmin)
admin.site.register(PointGroup,PointGroupAdmin)



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-10-03 19:22
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0019_invoice_buyerpayssalestax'),
        ('discounts', '0005_auto_20170830_1159'),
    ]

    operations = [
        migrations.CreateModel(
            name='CustomerDiscount',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('customer', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.Customer', verbose_name='Customer')),
                ('discountCombo', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='discounts.DiscountCombo', verbose_name='Discount')),
            ],
            options={
                'verbose_name_plural': 'Customer-specific discount restrictions',
                'verbose_name': 'Customer-specific discount restriction',
            },
        ),
    ]



from django.conf.urls import url

from .feeds import LatestNewsFeed
from .views import NewsView

urlpatterns = [
 	# This shows the latest news, paginated
 	url(r'^$',NewsView.as_view(), name='news'),

    # This is the latest news feed
    url(r'^feed/$',LatestNewsFeed(), name='news_feed'),
]



from django.core.urlresolvers import reverse
from django.core.exceptions import ObjectDoesNotExist, ValidationError
from django.contrib import messages
from django.db.models import Q
from django.http import HttpResponseRedirect, Http404
from django.views.generic import FormView, RedirectView, TemplateView
from django.utils.translation import ugettext_lazy as _
from django.utils import timezone
from django.utils.dateparse import parse_datetime

from braces.views import UserFormKwargsMixin
import logging
from allauth.account.forms import LoginForm, SignupForm
from datetime import timedelta
import json

from .models import Event, Series, PublicEvent, TemporaryRegistration, TemporaryEventRegistration, Invoice, CashPaymentRecord
from .forms import ClassChoiceForm, RegistrationContactForm, DoorAmountForm
from .constants import getConstant, REG_VALIDATION_STR
from .signals import post_student_info, request_discounts, apply_discount, apply_addons, apply_price_adjustments
from .mixins import FinancialContextMixin


# Define logger for this file
logger = logging.getLogger(__name__)


class RegistrationOfflineView(TemplateView):
    '''
    If registration is offline, just say so.
    '''
    template_name = 'core/registration/registration_offline.html'


class ClassRegistrationReferralView(RedirectView):

    def get(self,request,*args,**kwargs):

        # Always redirect to the classes page
        self.url = reverse('registration')

        # Voucher IDs are used for the referral program.
        # Marketing IDs are used for tracking click-through registrations.
        # They are put directly into session data immediately.
        voucher_id = kwargs.pop('voucher_id',None)
        marketing_id = kwargs.pop('marketing_id',None)

        if marketing_id or voucher_id:
            ''' Put these things into the session data. '''
            regSession = self.request.session.get(REG_VALIDATION_STR, {})
            regSession['voucher_id'] = voucher_id or regSession.get('voucher_id',None)
            regSession['marketing_id'] = marketing_id or regSession.get('marketing_id',None)
            self.request.session[REG_VALIDATION_STR] = regSession

        return super(ClassRegistrationReferralView,self).get(request,*args,**kwargs)


class ClassRegistrationView(FinancialContextMixin, FormView):
    '''
    This is the main view that is called from the class registration page.y
    '''
    form_class = ClassChoiceForm
    template_name = 'core/registration/event_registration.html'

    def get(self, request, *args, **kwargs):
        ''' Check that registration is online before proceeding '''
        regonline = getConstant('registration__registrationEnabled')
        if not regonline:
            return HttpResponseRedirect(reverse('registrationOffline'))

        return super(ClassRegistrationView,self).get(request,*args,**kwargs)

    def get_context_data(self,**kwargs):
        ''' Add the event and series listing data '''
        context = self.get_listing()
        context.update(kwargs)

        return super(ClassRegistrationView,self).get_context_data(**context)

    def get_form_kwargs(self, **kwargs):
        ''' Tell the form which fields to render '''
        kwargs = super(ClassRegistrationView, self).get_form_kwargs(**kwargs)
        kwargs['user'] = self.request.user if hasattr(self.request,'user') else None

        listing = self.get_listing()

        kwargs.update({
            'openEvents': listing['openEvents'],
            'closedEvents': listing['closedEvents'],
        })
        return kwargs

    def form_valid(self,form):
        '''
        If the form is valid, pass its contents on to the next view.  In order to permit the registration
        form to be overridden flexibly, but without permitting storage of arbitrary data keys that could
        lead to potential security issues, a form class for this view can optionally specify a list of
        keys that are permitted.  If no such list is specified as instance.permitted_event_keys, then
        the default list are used.
        '''
        regSession = self.request.session.get(REG_VALIDATION_STR, {})

        # The session expires after a period of inactivity that is specified in preferences.
        expiry = timezone.now() + timedelta(minutes=getConstant('registration__sessionExpiryMinutes'))
        permitted_keys = getattr(form,'permitted_event_keys',['role',])

        try:
            event_listing = {
                int(key.split("_")[-1]): {k:v for k,v in json.loads(value[0]).items() if k in permitted_keys}
                for key,value in form.cleaned_data.items() if 'event' in key and value
            }
            non_event_listing = {key: value for key,value in form.cleaned_data.items() if 'event' not in key}
        except (ValueError, TypeError) as e:
            form.add_error(None,ValidationError(_('Invalid event information passed.'),code='invalid'))
            return super(ClassRegistrationView,self).form_invalid(form)

        associated_events = Event.objects.filter(id__in=[k for k in event_listing.keys()])

        # Include the submission user if the user is authenticated
        if self.request.user.is_authenticated:
            submissionUser = self.request.user
        else:
            submissionUser = None

        reg = TemporaryRegistration(
            submissionUser=submissionUser,dateTime=timezone.now(),
            payAtDoor=non_event_listing.pop('payAtDoor',False),
            expirationDate=expiry,
        )

        # Anything passed by this form that is not an Event field (any extra fields) are
        # passed directly into the TemporaryRegistration's data.
        reg.data = non_event_listing or {}

        if regSession.get('marketing_id'):
            reg.data.update({'marketing_id': regSession.pop('marketing_id',None)})

        eventRegs = []
        grossPrice = 0

        for key,value in event_listing.items():
            this_event = associated_events.get(id=key)

            # Check if registration is still feasible based on both completed registrations
            # and registrations that are not yet complete
            this_role_id = value.get('role',None) if 'role' in permitted_keys else None
            soldOut = this_event.soldOutForRole(role=this_role_id,includeTemporaryRegs=True)

            if soldOut:
                if self.request.user.has_perm('core.override_register_soldout'):
                    # This message will be displayed on the Step 2 page by default.
                    messages.warning(self.request,_(
                        'Registration for \'%s\' is sold out. Based on your user permission level, ' % this_event.name +
                        'you may proceed with registration.  However, if you do not wish to exceed ' +
                        'the listed capacity of the event, please do not proceed.'
                    ))
                else:
                    # For users without permissions, don't allow registration for sold out things
                    # at all.
                    form.add_error(None, ValidationError(_('Registration for "%s" is tentatively sold out while others complete their registration.  Please try again later.' % this_event.name),code='invalid'))
                    return super(ClassRegistrationView,self).form_invalid(form)

            dropInList = [int(k.split("_")[-1]) for k,v in value.items() if k.startswith('dropin_') and v is True]

            # If nothing is sold out, then proceed to create a TemporaryRegistration and
            # TemporaryEventRegistration objects for the items selected by this form.  The
            # expiration date is set to be identical to that of the session.

            logger.debug('Creating temporary event registration for: %s' % key)
            if len(dropInList) > 0:
                tr = TemporaryEventRegistration(
                    event=this_event, dropIn=True, price=this_event.getBasePrice(dropIns=len(dropInList))
                )
            else:
                tr = TemporaryEventRegistration(
                    event=this_event, price=this_event.getBasePrice(payAtDoor=reg.payAtDoor), role_id=this_role_id
                )
            # If it's possible to store additional data and such data exist, then store them.
            tr.data = {k: v for k,v in value.items() if k in permitted_keys and k != 'role'}
            eventRegs.append(tr)
            grossPrice += tr.price

        # If we got this far with no issues, then save
        reg.priceWithDiscount = grossPrice
        reg.save()
        for er in eventRegs:
            er.registration = reg
            er.save()

        regSession["temporaryRegistrationId"] = reg.id
        regSession["temporaryRegistrationExpiry"] = expiry.strftime('%Y-%m-%dT%H:%M:%S%z')
        self.request.session[REG_VALIDATION_STR] = regSession
        return HttpResponseRedirect(self.get_success_url())

    def get_success_url(self):
        return reverse('getStudentInfo')

    def get_allEvents(self):
        '''
        Splitting this method out to get the set of events to filter allows
        one to subclass for different subsets of events without copying other
        logic
        '''

        if not hasattr(self,'allEvents'):
            timeFilters = {'endTime__gte': timezone.now()}
            if getConstant('registration__displayLimitDays') or 0 > 0:
                timeFilters['startTime__lte'] = timezone.now() + timedelta(days=getConstant('registration__displayLimitDays'))

            # Get the Event listing here to avoid duplicate queries
            self.allEvents = Event.objects.filter(
                **timeFilters
            ).filter(
                Q(instance_of=PublicEvent) |
                Q(instance_of=Series)
            ).exclude(
                Q(status=Event.RegStatus.hidden) |
                Q(status=Event.RegStatus.regHidden) |
                Q(status=Event.RegStatus.linkOnly)
            ).order_by('year','month','startTime')

        return self.allEvents

    def get_listing(self):
        '''
        This function gets all of the information that we need to either render or
        validate the form.  It is structured to avoid duplicate DB queries
        '''
        if not hasattr(self,'listing'):
            allEvents = self.get_allEvents()

            openEvents = allEvents.filter(registrationOpen=True)
            closedEvents = allEvents.filter(registrationOpen=False)

            publicEvents = allEvents.instance_of(PublicEvent)
            allSeries = allEvents.instance_of(Series)

            self.listing = {
                'allEvents': allEvents,
                'openEvents': openEvents,
                'closedEvents': closedEvents,
                'publicEvents': publicEvents,
                'allSeries': allSeries,
                'regOpenEvents': publicEvents.filter(registrationOpen=True).filter(
                    Q(publicevent__category__isnull=True) | Q(publicevent__category__separateOnRegistrationPage=False)
                ),
                'regClosedEvents': publicEvents.filter(registrationOpen=False).filter(
                    Q(publicevent__category__isnull=True) | Q(publicevent__category__separateOnRegistrationPage=False)
                ),
                'categorySeparateEvents': publicEvents.filter(
                    publicevent__category__separateOnRegistrationPage=True
                ).order_by('publicevent__category'),
                'regOpenSeries': allSeries.filter(registrationOpen=True).filter(
                    Q(series__category__isnull=True) | Q(series__category__separateOnRegistrationPage=False)
                ),
                'regClosedSeries': allSeries.filter(registrationOpen=False).filter(
                    Q(series__category__isnull=True) | Q(series__category__separateOnRegistrationPage=False)
                ),
                'categorySeparateSeries': allSeries.filter(
                    series__category__separateOnRegistrationPage=True
                ).order_by('series__category'),
            }
        return self.listing


class SingleClassRegistrationView(ClassRegistrationView):
    '''
    This view is called only via a link, and it allows a person to register for a single
    class without seeing all other classes.
    '''
    template_name = 'core/registration/single_event_registration.html'

    def get_allEvents(self):
        try:
            self.allEvents = Event.objects.filter(uuid=self.kwargs.get('uuid','')).exclude(status=Event.RegStatus.hidden)
        except ValueError:
            raise Http404()

        if not self.allEvents:
            raise Http404()

        return self.allEvents


class RegistrationSummaryView(UserFormKwargsMixin, FinancialContextMixin, FormView):
    template_name = 'core/registration_summary.html'
    form_class = DoorAmountForm

    def dispatch(self,request,*args,**kwargs):
        ''' Always check that the temporary registration has not expired '''
        regSession = self.request.session.get(REG_VALIDATION_STR,{})

        if not regSession:
            return HttpResponseRedirect(reverse('registration'))

        try:
            reg = TemporaryRegistration.objects.get(
                id=self.request.session[REG_VALIDATION_STR].get('temporaryRegistrationId')
            )
        except ObjectDoesNotExist:
            messages.error(request,_('Invalid registration identifier passed to summary view.'))
            return HttpResponseRedirect(reverse('registration'))

        expiry = parse_datetime(
            self.request.session[REG_VALIDATION_STR].get('temporaryRegistrationExpiry',''),
        )
        if not expiry or expiry < timezone.now():
            messages.info(request,_('Your registration session has expired. Please try again.'))
            return HttpResponseRedirect(reverse('registration'))

        # If OK, pass the registration and proceed
        kwargs.update({
            'reg': reg,
        })
        return super(RegistrationSummaryView,self).dispatch(request, *args, **kwargs)

    def get(self,request,*args,**kwargs):
        reg = kwargs.get('reg')

        initial_price = sum([x.price for x in reg.temporaryeventregistration_set.all()])

        # If the discounts app is enabled, then the return value to this signal
        # will contain information on the discounts to be applied, as well as
        # the total price of discount-ineligible items to be added to the
        # price.  These should be in the form of a named tuple such as the
        # DiscountApplication namedtuple defined in the discounts app, with
        # 'items' and 'ineligible_total' keys.
        discount_responses = request_discounts.send(
            sender=RegistrationSummaryView,
            registration=reg,
        )
        discount_responses = [x[1] for x in discount_responses if len(x) > 1 and x[1]]

        # This signal handler is designed to handle a single non-null response,
        # and that response must be in the form of a list of namedtuples, each
        # with a with a code value, a net_price value, and a discount_amount value
        # (as with the DiscountInfo namedtuple provided by the DiscountCombo class). If more
        # than one response is received, then the one with the minumum net price is applied
        discount_codes = []
        discounted_total = initial_price
        total_discount_amount = 0

        try:
            if discount_responses:
                discount_responses.sort(key=lambda k: min([getattr(x,'net_price',initial_price) for x in k.items] + [initial_price]) if k and hasattr(k,'items') else initial_price)
                discount_codes = getattr(discount_responses[0],'items',[])
                if discount_codes:
                    discounted_total = min([getattr(x,'net_price',initial_price) for x in discount_codes]) + getattr(discount_responses[0],'ineligible_total',0)
                    total_discount_amount = initial_price - discounted_total
        except (IndexError, TypeError) as e:
            logger.error('Error in applying discount responses: %s' % e)

        for discount in discount_codes:
            apply_discount.send(
                sender=RegistrationSummaryView,
                discount=discount.code,
                discount_amount=discount.discount_amount,
                registration=reg,
            )

        # Get any free add-on items that should be applied
        addon_responses = apply_addons.send(
            sender=RegistrationSummaryView,
            registration=reg
        )
        addons = []
        for response in addon_responses:
            try:
                if response[1]:
                    addons += list(response[1])
            except (IndexError, TypeError) as e:
                logger.error('Error in applying addons: %s' % e)

        # The return value to this signal should contain any adjustments that
        # need to be made to the price (e.g. from vouchers if the voucher app
        # is installed)
        adjustment_responses = apply_price_adjustments.send(
            sender=RegistrationSummaryView,
            registration=reg,
            initial_price=discounted_total
        )

        adjustment_list = []
        adjustment_amount = 0

        for response in adjustment_responses:
            adjustment_list += response[1][0]
            adjustment_amount += response[1][1]

        # Save the discounted price to the database
        total = discounted_total - adjustment_amount
        reg.priceWithDiscount = total
        reg.save()

        # Update the session key to keep track of this registration
        regSession = request.session[REG_VALIDATION_STR]
        regSession["temp_reg_id"] = reg.id
        if discount_codes:
            regSession['discount_codes'] = [(x.code.name, x.code.pk, x.discount_amount) for x in discount_codes]
        regSession['total_discount_amount'] = total_discount_amount
        regSession['addons'] = addons
        regSession['voucher_names'] = adjustment_list
        regSession['total_voucher_amount'] = adjustment_amount
        request.session[REG_VALIDATION_STR] = regSession

        return super(RegistrationSummaryView,self).get(request,*args,**kwargs)

    def get_context_data(self,**kwargs):
        ''' Pass the initial kwargs, then update with the needed registration info. '''
        context_data = super(RegistrationSummaryView,self).get_context_data(**kwargs)

        regSession = self.request.session[REG_VALIDATION_STR]
        reg_id = regSession["temp_reg_id"]
        reg = TemporaryRegistration.objects.get(id=reg_id)

        discount_codes = regSession.get('discount_codes',None)
        discount_amount = regSession.get('total_discount_amount',0)
        voucher_names = regSession.get('voucher_names',[])
        total_voucher_amount = regSession.get('total_voucher_amount',0)
        addons = regSession.get('addons',[])

        if reg.priceWithDiscount == 0:
            # Create a new Invoice if one does not already exist.
            new_invoice = Invoice.get_or_create_from_registration(reg,status=Invoice.PaymentStatus.paid)
            new_invoice.processPayment(0,0,forceFinalize=True)
            isFree = True
        else:
            isFree = False

        context_data.update({
            'registration': reg,
            "totalPrice": reg.totalPrice,
            'subtotal': reg.priceWithDiscount,
            'taxes': reg.addTaxes,
            "netPrice": reg.priceWithDiscountAndTaxes,
            "addonItems": addons,
            "discount_codes": discount_codes,
            "discount_code_amount": discount_amount,
            "voucher_names": voucher_names,
            "total_voucher_amount": total_voucher_amount,
            "total_discount_amount": discount_amount + total_voucher_amount,
            "currencyCode": getConstant('general__currencyCode'),
            'payAtDoor': reg.payAtDoor,
            'is_free': isFree,
        })

        if self.request.user:
            door_permission = self.request.user.has_perm('core.accept_door_payments')
            invoice_permission = self.request.user.has_perm('core.send_invoices')

            if door_permission or invoice_permission:
                context_data['form'] = DoorAmountForm(
                    user=self.request.user,
                    doorPortion=door_permission,
                    invoicePortion=invoice_permission,
                    payerEmail=reg.email,
                    discountAmount=max(reg.totalPrice - reg.priceWithDiscount,0),
                )

        return context_data

    def form_valid(self,form):
        regSession = self.request.session[REG_VALIDATION_STR]
        reg_id = regSession["temp_reg_id"]
        tr = TemporaryRegistration.objects.get(id=reg_id)

        # Create a new Invoice if one does not already exist.
        new_invoice = Invoice.get_or_create_from_registration(tr)

        if form.cleaned_data.get('paid'):
            logger.debug('Form is marked paid. Preparing to process payment.')

            amount = form.cleaned_data["amountPaid"]
            submissionUser = form.cleaned_data.get('submissionUser')
            receivedBy = form.cleaned_data.get('receivedBy')
            payerEmail = form.cleaned_data.get('cashPayerEmail')

            this_cash_payment = CashPaymentRecord.objects.create(
                invoice=new_invoice,
                submissionUser=submissionUser,
                amount=amount,
                payerEmail=payerEmail,
                collectedByUser=receivedBy,
                status=CashPaymentRecord.PaymentStatus.needsCollection,
            )

            # Process payment, but mark cash payment as needing collection from
            # the user who processed the registration and collected it.
            new_invoice.processPayment(
                amount,0,
                paidOnline=False,
                methodName='Cash',
                methodTxn='CASHPAYMENT_%s' % (this_cash_payment.recordId),
                submissionUser=submissionUser,
                collectedByUser=receivedBy,
                status=Invoice.PaymentStatus.needsCollection,
                forceFinalize=True,
            )
        elif form.cleaned_data.get('invoiceSent'):
            # Do not finalize this registration, but set the expiration date
            # on the TemporaryRegistration such that it will not be deleted
            # until after the last series ends, in case this person does not make
            # a payment right away.  This will also hold this individual's spot
            # in anything for which they have registered indefinitely.
            payerEmail = form.cleaned_data['invoicePayerEmail']
            tr.expirationDate = tr.lastEndTime
            tr.save()
            new_invoice.sendNotification(payerEmail=payerEmail,newRegistration=True)

        return HttpResponseRedirect(reverse('registration'))


class StudentInfoView(FormView):
    '''
    This page displays a preliminary total of what is being signed up for, and it also
    collects customer information, either by having the user sign in in an Ajax view, or by
    manually entering the information.  When the form is submitted, the view just passes
    everything into the session data and continues on to the next step.  To add additional
    fields to this form, or to modify existing fields, just override the form class to
    a form that adds/modifies whatever fields you would like.
    '''
    form_class = RegistrationContactForm
    template_name = 'core/student_info_form.html'

    def dispatch(self, request, *args, **kwargs):
        '''
        Require session data to be set to proceed, otherwise go back to step 1.
        Because they have the same expiration date, this also implies that the
        TemporaryRegistration object is not yet expired.
        '''
        if REG_VALIDATION_STR not in request.session:
            return HttpResponseRedirect(reverse('registration'))

        try:
            self.temporaryRegistration = TemporaryRegistration.objects.get(
                id=self.request.session[REG_VALIDATION_STR].get('temporaryRegistrationId')
            )
        except ObjectDoesNotExist:
            messages.error(request,_('Invalid registration identifier passed to sign-up form.'))
            return HttpResponseRedirect(reverse('registration'))

        expiry = parse_datetime(
            self.request.session[REG_VALIDATION_STR].get('temporaryRegistrationExpiry',''),
        )
        if not expiry or expiry < timezone.now():
            messages.info(request,_('Your registration session has expired. Please try again.'))
            return HttpResponseRedirect(reverse('registration'))

        return super(StudentInfoView,self).dispatch(request,*args,**kwargs)

    def get_context_data(self, **kwargs):
        context_data = super(StudentInfoView,self).get_context_data(**kwargs)
        reg = self.temporaryRegistration

        initial_price = sum([x.price for x in reg.temporaryeventregistration_set.all()])

        # If the discounts app is enabled, then the return value to this signal
        # will contain information on the discounts to be applied, as well as
        # the total price of discount-ineligible items to be added to the
        # price.  These should be in the form of a named tuple such as the
        # DiscountApplication namedtuple defined in the discounts app, with
        # 'items' and 'ineligible_total' keys.
        discount_responses = request_discounts.send(
            sender=StudentInfoView,
            registration=reg,
        )
        discount_responses = [x[1] for x in discount_responses if len(x) > 1 and x[1]]

        # This signal handler is designed to handle a single non-null response,
        # and that response must be in the form of a list of namedtuples, each
        # with a with a code value, a net_price value, and a discount_amount value
        # (as with the DiscountInfo namedtuple provided by the DiscountCombo class). If more
        # than one response is received, then the one with the minumum net price is applied
        discount_codes = []
        discounted_total = initial_price
        total_discount_amount = 0

        try:
            if discount_responses:
                discount_responses.sort(key=lambda k: min([getattr(x,'net_price',initial_price) for x in k.items] + [initial_price]) if k and hasattr(k,'items') else initial_price)
                discount_codes = getattr(discount_responses[0],'items',[])
                if discount_codes:
                    discounted_total = min([getattr(x,'net_price',initial_price) for x in discount_codes]) + getattr(discount_responses[0],'ineligible_total',0)
                    total_discount_amount = initial_price - discounted_total
        except (IndexError, TypeError) as e:
            logger.error('Error in applying discount responses: %s' % e)

        # Get any free add-on items that should be applied
        addon_responses = apply_addons.send(
            sender=StudentInfoView,
            registration=reg
        )
        addons = []
        for response in addon_responses:
            try:
                if response[1]:
                    addons += list(response[1])
            except (IndexError, TypeError) as e:
                logger.error('Error in applying addons: %s' % e)

        context_data.update({
            'reg': reg,
            'payAtDoor': reg.payAtDoor,
            'currencySymbol': getConstant('general__currencySymbol'),
            'subtotal': initial_price,
            'addonItems': addons,
            'discount_codes': discount_codes,
            'discount_code_amount': total_discount_amount,
            'discounted_subtotal': discounted_total,
        })

        if reg.payAtDoor or self.request.user.is_authenticated or not getConstant('registration__allowAjaxSignin'):
            context_data['show_ajax_form'] = False
        else:
            # Add a login form and a signup form
            context_data.update({
                'show_ajax_form': True,
                'login_form': LoginForm(),
                'signup_form': SignupForm(),
            })

        return context_data

    def get_form_kwargs(self, **kwargs):
        ''' Pass along the request data to the form '''
        kwargs = super(StudentInfoView, self).get_form_kwargs(**kwargs)
        kwargs['request'] = self.request
        kwargs['registration'] = self.temporaryRegistration
        return kwargs

    def get_success_url(self):
        return reverse('showRegSummary')

    def form_valid(self,form):
        '''
        Even if this form is valid, the handlers for this form may have added messages
        to the request.  In that case, then the page should be handled as if the form
        were invalid.  Otherwise, update the session data with the form data and then
        move to the next view
        '''
        reg = self.temporaryRegistration

        # The session expires after a period of inactivity that is specified in preferences.
        expiry = timezone.now() + timedelta(minutes=getConstant('registration__sessionExpiryMinutes'))
        self.request.session[REG_VALIDATION_STR]["temporaryRegistrationExpiry"] = \
            expiry.strftime('%Y-%m-%dT%H:%M:%S%z')
        self.request.session.modified = True

        # Update the expiration date for this registration, and pass in the data from
        # this form.
        reg.expirationDate = expiry
        reg.firstName = form.cleaned_data.pop('firstName')
        reg.lastName = form.cleaned_data.pop('lastName')
        reg.email = form.cleaned_data.pop('email')
        reg.phone = form.cleaned_data.pop('phone', None)
        reg.student = form.cleaned_data.pop('student',False)
        reg.comments = form.cleaned_data.pop('comments',None)
        reg.howHeardAboutUs = form.cleaned_data.pop('howHeardAboutUs',None)

        # Anything else in the form goes to the TemporaryRegistration data.
        reg.data.update(form.cleaned_data)
        reg.save()

        # This signal (formerly the post_temporary_registration signal) allows
        # vouchers to be applied temporarily, and it can be used for other tasks
        post_student_info.send(sender=StudentInfoView,registration=reg)
        return HttpResponseRedirect(self.get_success_url())  # Redirect after POST
